"""
NRV-fascicles
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import os

import matplotlib.pyplot as plt
import numpy as np

from ..backend.file_handler import *
from ..backend.log_interface import pass_info, rise_warning
from ..backend.MCore import MCH, synchronize_processes
from ..backend.NRV_Simulable import NRV_simulable, sim_results
from ..fmod.extracellular import *
from ..fmod.recording import *
from ..utils.cell.CL_postprocessing import *
from .axons import *
from .fascicle_generator import *
from .myelinated import *
from .unmyelinated import *



# Parallel computing options
if not MCH.is_alone():
    fg_verbose = False
    if MCH.do_master_only_work():
        fg_verbose = True

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()

OTF_PP_path = os.environ["NRVPATH"] + "/_misc/OTF_PP/"
OTF_PP_library = os.listdir(OTF_PP_path)

#####################
## Fasscicle class ##
#####################


def is_fascicle(object):
    """
    check if an object is a fascicle, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a fascicle object
    """
    return isinstance(object, fascicle)


class fascicle(NRV_simulable):
    """
    Class for Fascicle, defined as a group of axons near one to the other in the same Perineurium Sheath. All axons are independant of each other, no ephaptic coupling.
    """

    def __init__(self, ID=0, **kwargs):
        """
        Instrantation of a Fascicle

        Parameters
        ----------
        dt              : float
            simulation time stem for Neuron (ms), by default 1us
        Nseg_per_sec    : float
            number of segment per section in Neuron. If set to 0, the number of segment per section is calculated with the d-lambda rule
        freq            : float
            frequency of the d-lambda rule (Hz), by default 100Hz
        freq_min        : float
            minimum frequency for the d-lambda rule when meshing is irregular, 0 for regular meshing
        v_init          : float
            initial value for the membrane voltage (mV), specify None for automatic model choice of v_init
        T               : float
            temperature (C), specify None for automatic model choice of temperature
        ID              : int
            axon ID, by default set to 0
        threshold       : int
            membrane voltage threshold for spike detection (mV), by default -40mV
        """
        super().__init__()

        # to add to a fascicle/nerve common mother class
        self.save_path=""
        self.verbose=False
        self.postproc_script="default"
        self.return_parameters_only = True
        self.loaded_footprints = True
        self.save_results = True

        self.config_filename = ""
        self.type = "fascicle"
        self.ID = ID
        self.type = None
        self.L = None
        self.D = None
        # geometric properties
        self.y_grav_center = 0
        self.z_grav_center = 0
        self.N_vertices = 0
        self.y_vertices = np.array([])
        self.z_vertices = np.array([])
        self.A = 0
        # axonal content
        self.N = 0
        self.axons_diameter = np.array([])
        self.axons_type = np.array([])
        self.axons_y = np.array([])
        self.axons_z = np.array([])
        self.NoR_relative_position = np.array([])
        # Axons objects default parameters
        self.unmyelinated_param = {
            "model": "Rattay_Aberham",
            "dt": 0.001,
            "Nrec": 0,
            "Nsec": 1,
            "Nseg_per_sec": 100,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "v_init": None,
            "T": None,
            "threshold": -40,
        }

        self.myelinated_param = {
            "model": "MRG",
            "dt": 0.001,
            "node_shift": 0,
            "Nseg_per_sec": 3,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "rec": "nodes",
            "T": None,
            "threshold": -40,
        }
        self.set_axons_parameters(**kwargs)

        # extra-cellular stimulation
        self.extra_stim = None
        self.footprints = {}
        self.is_footprinted = False
        # intra-cellular stimulation
        self.N_intra = 0
        self.intra_stim_position = []
        self.intra_stim_t_start = []
        self.intra_stim_duration = []
        self.intra_stim_amplitude = []
        self.intra_stim_ON = []
        ## recording mechanism
        self.record = False
        self.recorder = None
        # Simulation status
        self.is_simulated = False
        self.processes_status = ["preparing" for _ in range(MCH.size)]

    ## save/load methods
    def save(
        self,
        fname="fascicle.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        save=True,
        _fasc_save=True,
        blacklist=[],
    ):
        """
        Save a fascicle in a json file

        Parameters
        ----------
        fname :             str
            name of the file to save the fascicle
        extracel_context:   bool
            if True, add the extracellular context to the saving
        intracel_context:   bool
            if True, add the intracellular context to the saving
        rec_context:        bool
            if True, add the recording context to the saving
        blacklist:          list[str]
            key to exclude from saving
        save:               bool
            if false only return the dictionary

        Returns
        -------
        key_dict:           dict
            Pyhton dictionary containing all the fascicle information
        """
        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += ["extra_stim", "footprints", "is_footprinted"]

        if not rec_context:
            bl += ["record", "recorder"]

        to_save = (save and _fasc_save) and MCH.do_master_only_work()
        return super().save(fname=fname, save=to_save, blacklist=bl)

    def load(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Load a fascicle configuration from a json file

        Parameters
        ----------
        fname           : str
            path to the json file describing a fascicle
        extracel_context: bool
            if True, load the extracellular context as well
        intracel_context: bool
            if True, load the intracellular context as well
        rec_context: bool
            if True, load the recording context as well
        blacklist       : list[str]
            key to exclude from loading
        """
        if type(data) == str:
            key_dic = json_load(data)
        else:
            key_dic = data

        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += [
                "extra_stim",
                "footprints",
                "myelinated_nseg_per_sec",
                "unmyelinated_nseg",
                "is_footprinted",
            ]

        if not rec_context:
            bl += ["record", "recorder"]

        super().load(data=key_dic, blacklist=bl)
        if "unmyelinated_param" not in key_dic:
            rise_warning("Deprecated fascicle file")
            self.set_axons_parameters(key_dic)

    def save_fascicle_configuration(
        self, fname, extracel_context=False, intracel_context=False, rec_context=False
    ):
        rise_warning(
            "save_fascicle_configuration is a deprecated method use save instead"
        )
        self.save(
            fname=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    def load_fascicle_configuration(
        self, fname, extracel_context=False, intracel_context=False, rec_context=False
    ):
        rise_warning(
            "load_fascicle_configuration is a deprecated method use load instead"
        )
        self.load(
            data=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    ## Fascicle property method
    def set_ID(self, ID):
        """
        set the ID of the fascicle

        Parameters
        ----------
        ID : int
            ID number of the fascicle
        """
        self.ID = ID

    def define_length(self, L):
        """
        set the length over the x axis of the fascicle

        Parameters
        ----------
        L   : float
            length of the fascicle in um
        """
        if self.extra_stim is not None or self.N_intra > 0:
            rise_warning(
                "Modifying length of a fascicle with extra or intra cellular context can lead to error"
            )
        self.L = L
        self.set_axons_parameters(unmyelinated_nseg=self.L // 25)

    ## generate stereotypic Fascicle
    def define_circular_contour(self, D, y_c=None, z_c=None, N_vertices=100):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        D           : float
            diameter of the circular fascicle contour, in um
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        N_vertices  : int
            Number of vertice in the compute the contour
        """
        self.type = "Circular"
        self.D = D
        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        self.N_vertices = N_vertices
        theta = np.linspace(-np.pi, np.pi, num=N_vertices)
        self.y_vertices = self.y_grav_center + (D / 2) * np.cos(theta)
        self.z_vertices = self.z_grav_center + (D / 2) * np.sin(theta)
        self.A = np.pi * (D / 2) ** 2

    def get_circular_contour(self):
        """
        Returns the properties of the fascicle contour considered as a circle (y and z center and diameter)

        Returns
        -------
        D : float
            diameter of the contour, in um. Set to 0 if not applicable
        y : float
            y position of the contour center, in um
        z : float
            z position of the contour center, in um
        """
        y = self.y_grav_center
        z = self.z_grav_center
        if self.y_vertices == np.array([]) and self.D is None:
            D = 0
        elif self.D is not None:
            D = self.D
        else:
            y = (np.amax(self.y_vertices) - np.amin(self.y_vertices)) / 2
            z = (np.amax(self.z_vertices) - np.amin(self.z_vertices)) / 2
            D = np.abs(np.amax(self.y_vertices) - np.amin(self.y_vertices))
        return D, y, z

    def fit_circular_contour(self, y_c=None, z_c=None, Delta=0.1, N_vertices=100):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        Delta       : float
            distance between farest axon and contour, in um
        N_vertices  : int
            Number of vertice in the compute the contour
        """
        N_axons = len(self.axons_diameter)
        D = 2 * Delta

        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        if N_axons == 0:
            pass_info("No axon to fit fascicul diameter set to " + str(D) + "um")
        else:
            for axon in range(N_axons):
                dist_max = (
                    self.axons_diameter[axon] / 2
                    + (
                        (self.y_grav_center - self.axons_y[axon]) ** 2
                        + (self.z_grav_center - self.axons_z[axon]) ** 2
                    )
                    ** 0.5
                )
                D = max(D, 2 * (dist_max + Delta))
        self.define_circular_contour(D, y_c=None, z_c=None, N_vertices=N_vertices)

    def define_ellipsoid_contour(self, a, b, y_c=0, z_c=0, rotate=0):
        """
        Define ellipsoidal contour
        """
        pass

    ## generate Fascicle from histology
    def import_contour(self, smthing_else):
        """
        Define contour from a file
        """
        pass

    ## fill fascicle methods
    def fill(
        self,
        parallel=True,
        percent_unmyel=0.7,
        FVF=0.55,
        M_stat="Schellens_1",
        U_stat="Ochoa_U",
        ppop_fname=None,
        pop_fname=None,
    ):
        """
        Fill a geometricaly defined contour with axons

        Parameters
        ----------
        parallel        : bool
            if True, the generation process (quite long) is split over multiples cores, if False everything is perfrmed by the master.
        percent_unmyel  : float
            ratio of unmyelinated axons in the population. Should be between 0 and 1.
        FVF             : float
            Fiber Volume Fraction estimated for the area. By default set to 0.55
        M_stat          : str
            name of the statistic in the librairy or path to a new librairy in csv for myelinated diameters repartition
        U_stat          : str
            name of the statistic in the librairy or path to a new librairy in csv for unmyelinated diameters repartition
        ppop_fname      : str
            optional, if specified, name file to store the placed population generated
        pop_fname       : str
            optional, if specified, name file to store the population generated
        """
        #### AXON GENERATION: parallelization if resquested ####
        Area_to_fill = 0
        # Note: generate a bit too much axons just in case
        if self.type == "Circular":
            Area_to_fill = np.pi * (self.D / 2 + 28) ** 2
        if parallel:
            if MCH.do_master_only_work():
                pass_info("Generating axons")
            # split the generation over the N cores
            (
                partial_axons_diameter,
                partial_axons_type,
                M_diam_list,
                U_diam_list,
            ) = fill_area_with_axons(
                Area_to_fill / MCH.size,
                percent_unmyel=percent_unmyel,
                FVF=FVF,
                M_stat=M_stat,
                U_stat=U_stat,
            )
            # gather results
            axons_diameter = MCH.gather_jobs_as_array(partial_axons_diameter)
            axons_type = MCH.gather_jobs_as_array(partial_axons_type)
        else:
            if MCH.do_master_only_work():
                (
                    axons_diameter,
                    axons_type,
                    M_diam_list,
                    U_diam_list,
                ) = fill_area_with_axons(
                    Area_to_fill,
                    percent_unmyel=percent_unmyel,
                    FVF=FVF,
                    M_stat=M_stat,
                    U_stat=U_stat,
                )
        #### AXON PACKING: never parallel
        if MCH.do_master_only_work():
            N = len(axons_diameter)
            pass_info("\n ... " + str(N) + " axons generated")
            if pop_fname is not None:
                save_axon_population(
                    pop_fname, axons_diameter, axons_type, comment=None
                )
            pass_info("Axon Packing is starting")
            axons_y, axons_z, iteration, FVF_array, probed_iter = axon_packer(
                axons_diameter,
                Delta=0.1,
                y_gc=self.y_grav_center,
                z_gc=self.z_grav_center,
                max_iter=20000,
                monitor=False,
            )
            # check for remaining collisions and delete problematic axons
            axons_diameter, axons_type, axons_y, axons_z = delete_collisions(
                axons_diameter, axons_type, axons_y, axons_z
            )
            N = len(axons_diameter)
            pass_info("... Axon Packing done")
            # check if axons are inside the fascicle
            inside_axons = (
                np.power(axons_y - self.y_grav_center, 2)
                + np.power(axons_z - self.z_grav_center, 2)
                - np.power(np.ones(N) * (self.D / 2) - axons_diameter / 2, 2)
            )
            axons_to_keep = np.argwhere(inside_axons < 0)
            axons_diameter = axons_diameter[axons_to_keep]
            axons_type = axons_type[axons_to_keep]
            axons_y = axons_y[axons_to_keep]
            axons_z = axons_z[axons_to_keep]
            N = len(axons_diameter)
            # save the good population
            if ppop_fname is not None:
                save_placed_axon_population(
                    ppop_fname,
                    self.axons_diameter,
                    self.axons_type,
                    self.axons_y,
                    self.axons_z,
                    comment=None,
                )
        else:
            axons_diameter = None
            axons_type = None
            axons_y = None
            axons_z = None
            N = None
        ## BRODCASTING RESULTS TO ALL PARALLEL OBJECTS
        self.axons_diameter = MCH.master_broadcasts_array_to_all(
            axons_diameter
        ).flatten()
        self.axons_type = MCH.master_broadcasts_array_to_all(axons_type).flatten()
        self.axons_y = MCH.master_broadcasts_array_to_all(axons_y).flatten()
        self.axons_z = MCH.master_broadcasts_array_to_all(axons_z).flatten()
        self.N = MCH.master_broadcasts_array_to_all(N)

    def fill_with_population(
        self, axons_diameter, axons_type, Delta=0.1, ppop_fname=None, FVF=0.55
    ):
        """
        Fill a geometricaly defined contour with an already generated axon population

        Parameters
        ----------
        axons_diameters     : np.array
            Array  containing all the diameters of the axon population
        axons_type          : np.array
            Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
        ppop_fname      : str
            optional, if specified, name file to store the placed population generated
        FVF             : float
            Fiber Volume Fraction estimated for the area. By default set to 0.55
        """
        if MCH.do_master_only_work():
            ## check the population area
            total_ax_area = 0
            for diam in axons_diameter:
                total_ax_area += np.pi * ((diam / 2) ** 2)
            if self.A * FVF > total_ax_area:
                rise_warning(
                    "Warning: the specified population maybe too small to fill the current fascicle"
                )
            N = len(axons_diameter)
            pass_info("\n ... " + str(N) + " axons loaded")
            if ppop_fname is not None:
                save_axon_population(
                    ppop_fname, axons_diameter, axons_type, comment=None
                )
            pass_info("Axon Packing is starting")
            axons_y, axons_z, iteration, FVF_array, probed_iter = axon_packer(
                axons_diameter,
                Delta=Delta,
                y_gc=self.y_grav_center,
                z_gc=self.z_grav_center,
                max_iter=20000,
                monitor=False,
            )
            # check for remaining collisions and delete problematic axons
            axons_diameter, axons_type, axons_y, axons_z = delete_collisions(
                axons_diameter, axons_type, axons_y, axons_z
            )
            N = len(axons_diameter)
            pass_info("... Axon Packing done")
            # check if axons are inside the fascicle
            inside_axons = (
                np.power(axons_y - self.y_grav_center, 2)
                + np.power(axons_z - self.z_grav_center, 2)
                - np.power(np.ones(N) * (self.D / 2) - axons_diameter / 2, 2)
            )
            axons_to_keep = np.argwhere(inside_axons < 0)
            axons_diameter = axons_diameter[axons_to_keep]
            axons_type = axons_type[axons_to_keep]
            axons_y = axons_y[axons_to_keep]
            axons_z = axons_z[axons_to_keep]
            N = len(axons_diameter)
            # save the good population
            if ppop_fname is not None:
                save_placed_axon_population(
                    ppop_fname,
                    self.axons_diameter,
                    self.axons_type,
                    self.axons_y,
                    self.axons_z,
                    comment=None,
                )
        else:
            axons_diameter = None
            axons_type = None
            axons_y = None
            axons_z = None
            N = None
        ## BRODCASTING RESULTS TO ALL PARALLEL OBJECTS
        self.axons_diameter = MCH.master_broadcasts_array_to_all(
            axons_diameter
        ).flatten()
        self.axons_type = MCH.master_broadcasts_array_to_all(axons_type).flatten()
        self.axons_y = MCH.master_broadcasts_array_to_all(axons_y).flatten()
        self.axons_z = MCH.master_broadcasts_array_to_all(axons_z).flatten()
        self.N = MCH.master_broadcasts_array_to_all(N)

    def fill_with_placed_population(
        self,
        axons_diameter,
        axons_type,
        axons_y,
        axons_z,
        check_inside=True,
        check_collision=True,
        ppop_fname=None,
    ):
        """
        Fill a geometricaly defined contour with an already generated axon population

        Parameters
        ----------
        axons_diameters     : np.array
            Array  containing all the diameters of the axon population
        axons_type          : np.array
            Array containing a '1' value for indexes where the axon is myelinated (A-delta or not), else '0'
        axons_y             : np.array
            y coordinate of the axons, in um
        axons_z             : np.array
            z coordinate of the axons, in um
        check_inside        : bool
            if True the placed axons position are checked to ensure all remaining axons at the end are inside the fascicle
        check_collision     : bool
            if True, possible remaining collisions are check in the loaded population, and thiner axons are deleted
        ppop_fname          : str
            optional, if specified, name file to store the placed population generated
        """
        if MCH.do_master_only_work():
            N = len(axons_diameter)
            if check_collision:
                # check for remaining collisions and delete problematic axons
                axons_diameter, axons_type, axons_y, axons_z = delete_collisions(
                    axons_diameter, axons_type, axons_y, axons_z
                )
                N = len(axons_diameter)
            # check if axons are inside the fascicle
            if check_inside:
                inside_axons = (
                    np.power(axons_y - self.y_grav_center, 2)
                    + np.power(axons_z - self.z_grav_center, 2)
                    - np.power(np.ones(N) * (self.D / 2) - axons_diameter / 2, 2)
                )
                axons_to_keep = np.argwhere(inside_axons < 0)
                axons_diameter = axons_diameter[axons_to_keep]
                axons_type = axons_type[axons_to_keep]
                axons_y = axons_y[axons_to_keep]
                axons_z = axons_z[axons_to_keep]
                N = len(axons_diameter)
            # save the good population
            if ppop_fname is not None:
                save_placed_axon_population(
                    ppop_fname,
                    self.axons_diameter,
                    self.axons_type,
                    self.axons_y,
                    self.axons_z,
                    comment=None,
                )
        else:
            axons_diameter = None
            axons_type = None
            axons_y = None
            axons_z = None
            N = None
        ## BRODCASTING RESULTS TO ALL PARALLEL OBJECTS
        self.axons_diameter = MCH.master_broadcasts_array_to_all(
            axons_diameter
        ).flatten()
        self.axons_type = MCH.master_broadcasts_array_to_all(axons_type).flatten()
        self.axons_y = MCH.master_broadcasts_array_to_all(axons_y).flatten()
        self.axons_z = MCH.master_broadcasts_array_to_all(axons_z).flatten()
        self.N = MCH.master_broadcasts_array_to_all(N)

    ## Move methods
    def translate_axons(self, y, z):
        """
        Move axons only in a fascicle by group translation

        Parameters
        ----------
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        self.axons_y += y
        self.axons_z += z

    def translate_fascicle(self, y, z):
        """
        Translate a complete fascicle

        Parameters
        ----------
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        self.y_grav_center += y
        self.z_grav_center += z
        self.y_vertices += y
        self.z_vertices += z
        self.translate_axons(y, z)
        if self.extra_stim is not None:
            self.extra_stim.translate(y=y, z=z)
        if self.recorder is not None:
            self.recorder.translate(y=y, z=z)

    def rotate_axons(self, theta, y_c=0, z_c=0):
        """
        Move axons only in a fascicle by group rotation

        Parameters
        ----------
        theta   : float
            angular value of the translation, in rad
        y_c     : float
            y axis value of the rotation center in um, by default set to 0
        z_c     : float
            z axis value for the rotation center in um, by default set to 0
        """
        self.axons_y = (
            np.cos(theta) * (self.axons_y - y_c) - np.sin(theta) * (self.axons_z - z_c)
        ) + y_c
        self.axons_z = (
            np.sin(theta) * (self.axons_y - y_c) + np.cos(theta) * (self.axons_z - z_c)
        ) + z_c

    def rotate_fascicle(self, theta, y_c=0, z_c=0):
        """
        Rotate a complete fascicle

        Parameters
        ----------
        theta   : float
            angular value of the translation, in rad
        y_c     : float
            y axis value of the rotation center in um, by default set to 0
        z_c     : float
            z axis value for the rotation center in um, by default set to 0
        """
        self.y_grav_center = (
            np.cos(theta) * (self.y_grav_center - y_c)
            - np.sin(theta) * (self.z_grav_center - z_c)
        ) + y_c
        self.z_grav_center = (
            np.sin(theta) * (self.y_grav_center - y_c)
            + np.cos(theta) * (self.z_grav_center - z_c)
        ) + z_c
        self.y_vertices += (
            np.cos(theta) * (self.y_vertices - y_c)
            - np.sin(theta) * (self.z_vertices - z_c)
        ) + y_c
        self.z_vertices += (
            np.sin(theta) * (self.y_vertices - y_c)
            + np.cos(theta) * (self.z_vertices - z_c)
        ) + z_c
        self.rotate_axons(theta, y_c=y_c, z_c=z_c)

    def remove_axons_electrode_overlap(self, electrode):
        """
        Remove the axons that could overlap an electrode

        Parameters
        ----------
        electrode : object
            electrode instance, see electrodes for more details
        """
        y, z, D = 0, 0, 0
        if is_LIFE_electrode(electrode):
            D = electrode.D
            y = electrode.y
            z = electrode.z
        elif is_analytical_electrode(electrode):
            y = electrode.y
            z = electrode.z
        else:
            # CUFF electrodes should not affect intrafascicular state
            return 0
        # compute the distance of all axons to electrode
        D_vectors = np.sqrt((self.axons_y - y) ** 2 + (self.axons_z - z) ** 2) - (
            self.axons_diameter / 2 + D / 2
        )
        colapse = np.argwhere(D_vectors < 0)
        mask = np.ones(len(self.axons_diameter), dtype=bool)
        mask[colapse] = False
        # remove axons colliding the electrode
        if len(colapse) > 0:
            pass_info(
                "From Fascicle level: Electrode/Axons overlap, "
                + str(len(colapse))
                + " axons will be removed from the fascicle"
            )
        self.N -= len(colapse)
        pass_info(self.N, "axons remaining")
        self.axons_diameter = self.axons_diameter[mask]
        self.axons_type = self.axons_type[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]

    ## Axons related method
    def set_axons_parameters(
        self, unmyelinated_only=False, myelinated_only=False, **kwargs
    ):
        """
        set parameters of axons in the fascicle

        Parameters
        ----------
        unmyelinated_only:      bool
            if true modification only done for unmyelinated axons parameters, by default False
        myelinated_only:        bool
            if true modification only done for myelinated axons parameters, by default False
        kwargs:
            parameters to modify (see myelinated and unmyelinated)
        """
        ## Standard keys
        for key in kwargs:
            if "model" in key:
                if not myelinated_only and kwargs[key] in unmyelinated_models:
                    self.unmyelinated_param["model"] = kwargs[key]
                elif not unmyelinated_only and kwargs[key] in myelinated_models:
                    self.myelinated_param["model"] = kwargs[key]
                else:
                    rise_warning(kwargs[key], " is not an implemented axon model")
            else:
                if not myelinated_only and key in self.unmyelinated_param:
                    self.unmyelinated_param["model"] = kwargs[key]
                if not unmyelinated_only and key in self.myelinated_param:
                    self.myelinated_param["model"] = kwargs[key]

        ## Specific keys
        if "unmyelinated_nseg" in kwargs:
            self.unmyelinated_param["Nseg_per_sec"] = kwargs["unmyelinated_nseg"]
        if "myelinated_nseg_per_sec" in kwargs:
            self.myelinated_param["Nseg_per_sec"] = kwargs["myelinated_nseg_per_sec"]

    def get_axons_parameters(self, unmyelinated_only=False, myelinated_only=False):
        """
        get parameters of axons in the fascicle

        Parameters
        ----------
        unmyelinated_only:      bool
            modification will only
        myelinated_only:        bool
            modification will only
        """
        if unmyelinated_only:
            return self.unmyelinated_param
        if myelinated_only:
            return self.myelinated_param
        return self.unmyelinated_param, self.myelinated_param

    def remove_unmyelinated_axons(self):
        """
        Remove all unmyelinated fibers from the fascicle
        """
        mask = self.axons_type.astype(bool)
        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        self.N = len(self.axons_diameter)
        if len(self.NoR_relative_position) != 0:
            self.NoR_relative_position = self.NoR_relative_position[mask]

    def remove_myelinated_axons(self):
        """
        Remove all myelinated fibers from the
        """
        mask = np.invert(self.axons_type.astype(bool))
        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        self.N = len(self.axons_diameter)
        if len(self.NoR_relative_position) != 0:
            # almost useless but here for coherence
            self.NoR_relative_position = self.NoR_relative_position[mask]

    def remove_axons_size_threshold(self, d, min=True):
        """
        Remove fibers with diameters below/above a threshold
        """
        if min:
            mask = self.axons_diameter >= d
        else:
            mask = self.axons_diameter <= d

        self.axons_diameter = self.axons_diameter[mask]
        self.axons_y = self.axons_y[mask]
        self.axons_z = self.axons_z[mask]
        self.axons_type = self.axons_type[mask]
        self.N = len(self.axons_diameter)
        if len(self.NoR_relative_position) != 0:
            # almost useless but here for coherence
            self.NoR_relative_position = self.NoR_relative_position[mask]

    ## Representation methods
    def plot(
        self, fig, axes, contour_color="k", myel_color="r", unmyel_color="b", num=False
    ):
        """
        plot the fascicle in the Y-Z plane (transverse section)

        Parameters
        ----------
        fig     : matplotlib.figure
            figure to display the fascicle
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        contour_color   : str
            matplotlib color string applied to the contour. Black by default
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        num             : bool
            if True, the index of each axon is displayed on top of the circle
        """
        if MCH.do_master_only_work():
            ## plot contour
            axes.plot(
                self.y_vertices, self.z_vertices, linewidth=2, color=contour_color
            )
            ## plot axons
            circles = []
            for k in range(self.N):
                if self.axons_type[k] == 1:  # myelinated
                    circles.append(
                        plt.Circle(
                            (self.axons_y[k], self.axons_z[k]),
                            self.axons_diameter[k] / 2,
                            color=myel_color,
                            fill=True,
                        )
                    )
                else:
                    circles.append(
                        plt.Circle(
                            (self.axons_y[k], self.axons_z[k]),
                            self.axons_diameter[k] / 2,
                            color=unmyel_color,
                            fill=True,
                        )
                    )
            if self.extra_stim is not None:
                for electrode in self.extra_stim.electrodes:
                    if electrode.type == "LIFE":
                        circles.append(
                            plt.Circle(
                                (electrode.y, electrode.z),
                                electrode.D / 2,
                                color="gold",
                                fill=True,
                            )
                        )
                    elif not is_FEM_electrode(electrode):
                        axes.scatter(electrode.y, electrode.z, color="gold")
            for circle in circles:
                axes.add_patch(circle)
            if num:
                for k in range(self.N):
                    axes.text(self.axons_y[k], self.axons_z[k], str(k))

    def plot_x(
        self, fig, axes, myel_color="r", unmyel_color="b", Myelinated_model="MRG"
    ):
        """
        plot the fascicle's axons along Xline (longitudinal)

        Parameters
        ----------
        fig     : matplotlib.figure
            figure to display the fascicle
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        Myelinated_model : str
            model use for myelinated axon (use to calulated node position)
        """
        if MCH.do_master_only_work():
            if self.L is None or self.NoR_relative_position != []:
                drange = [
                    min(self.axons_diameter.flatten()),
                    max(self.axons_diameter.flatten()),
                ]
                polysize = np.poly1d(np.polyfit(drange, [0.5, 5], 1))
                for k in range(self.N):
                    relative_pos = self.NoR_relative_position[k]
                    d = round(self.axons_diameter.flatten()[k], 2)
                    if self.axons_type.flatten()[k] == 0.0:
                        color = unmyel_color
                        size = polysize(d)
                        axes.plot([0, self.L], np.ones(2) + k - 1, color=color, lw=size)
                    else:
                        color = myel_color
                        size = polysize(d)
                        axon = myelinated(
                            0,
                            0,
                            d,
                            self.L,
                            model=Myelinated_model,
                            node_shift=self.NoR_relative_position[k],
                        )
                        x_nodes = axon.x_nodes
                        node_number = len(x_nodes)
                        del axon
                        axes.plot([0, self.L], np.ones(2) + k - 1, color=color, lw=size)
                        axes.scatter(
                            x_nodes,
                            np.ones(node_number) + k - 1,
                            marker="x",
                            color=color,
                        )

                ## plot electrode(s) if existings
                if self.extra_stim is not None:
                    for electrode in self.extra_stim.electrodes:
                        if not is_FEM_electrode(electrode):
                            axes.plot(
                                electrode.x * np.ones(2), [0, self.N - 1], color="gold"
                            )
                axes.set_xlabel("x (um)")
                axes.set_ylabel("axon ID")
                axes.set_yticks(np.arange(self.N))
                axes.set_xlim(0, self.L)
                plt.tight_layout()

    ## Context handeling methods
    def clear_context(
        self, extracel_context=True, intracel_context=True, rec_context=True
    ):
        """
        Clear all stimulation and recording mecanism
        """
        if extracel_context:
            self.L = None
            # extra-cellular stimulation
            self.extra_stim = None
            self.footprints = {}
            self.is_footprinted = False
        if intracel_context:
            # intra-cellular stimulation
            self.N_intra = 0
            self.intra_stim_position = []
            self.intra_stim_t_start = []
            self.intra_stim_duration = []
            self.intra_stim_amplitude = []
            self.intra_stim_ON = []
        if rec_context:
            ## recording mechanism
            self.record = False
            self.recorder = None
        self.is_simulated = False

    def attach_extracellular_stimulation(self, stimulation):
        """
        attach a extracellular context of simulation for an axon.

        parameters:
        ----------
            stimulation  : stimulation object, see Extracellular.stimulation help for more details
        """
        if is_extra_stim(stimulation):
            self.extra_stim = stimulation
        # remove everlaping axons
        for electrode in stimulation.electrodes:
            self.remove_axons_electrode_overlap(electrode)

    def change_stimulus_from_elecrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        parameters:
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                new stimulus to set
        """
        if self.extra_stim is not None:
            self.extra_stim.change_stimulus_from_elecrode(ID_elec, stimulus)
        else:
            rise_warning("Cannot be changed: no extrastim in the fascicle")


    def insert_I_Clamp(self, position, t_start, duration, amplitude, ax_list=None):
        """
        Insert a IC clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the fascicle. Note that all thin myelinated and myelinated
            will be stimulated in the nearest node of Ranvier around the clamp specified position
        t_start     : float
            starting time, in ms
        duration    : float
            duration of the pulse, in ms
        amplitude   : float
            amplitude of the pulse (nA)
        ax_list     : list, array, np.array
            list of axons to insert the clamp on, if None, all axons are stimulated
        """
        self.intra_stim_position.append(position)
        self.intra_stim_t_start.append(t_start)
        self.intra_stim_duration.append(duration)
        self.intra_stim_amplitude.append(amplitude)
        self.intra_stim_ON.append(ax_list)
        self.N_intra += 1

    ## RECORDING MECHANIMS
    def attach_extracellular_recorder(self, rec):
        """
        attach an extracellular recorder to the axon

        Parameters
        ----------
        rec     : recorder object
            see Recording.recorder help for more details
        """
        if is_recorder(rec):
            self.record = True
            self.recorder = rec

    def shut_recorder_down(self):
        """
        Shuts down the recorder locally
        """
        self.record = False

    ## SIMULATION HANDLING
    def generate_random_NoR_position(self):
        """
        Generates radom Node of Ranvier shifts to prevent from axons with the same diamters to be aligned.
        """
        # also generated for unmyelinated but the meaningless value won't be used
        self.NoR_relative_position = np.random.uniform(
            low=0.0, high=1.0, size=len(self.axons_diameter)
        )

    def generate_ligned_NoR_position(self, x=0):
        """
        Generates Node of Ranvier shifts to aligned a node of each axon to x postition.

        Parameters
        ----------
        x    : float
            x axsis value (um) on which node are lined, by default 0
        """
        # also generated for unmyelinated but the meaningless value won't be used
        self.NoR_relative_position = []

        for i in range(self.N):
            if self.axons_type.flatten()[i] == 0.0:
                self.NoR_relative_position += [0.0]
            else:
                d = round(self.axons_diameter.flatten()[i], 2)
                node_length = get_MRG_parameters(d)[5]
                self.NoR_relative_position += [(x - 0.5) % node_length / node_length]
                # -0.5 to be at the node of Ranvier center as a node is 1um long

    def compute_electrodes_footprints(
        self, save_ftp_only=False, filename="electrodes_footprint.ftpt", **kwargs
    ):
        """
        get electrodes footprints on each segment of each axon

        Parameters
        ----------
        save_ftp_only        :bool
            if true save result in a .ftpt file
        filename    :str
            saving file name and path
        Unmyelinated_model  : str
            model for unmyelinated fibers, by default 'Rattay_Aberham'
        Adelta_model        : str
            model for A-delta thin myelinated fibers, by default'extended_Gaines'
        Myelinated_model    : str
            model for myelinated fibers, by default 'MRG'
        myelinated_nseg_per_sec        : int
            number of segment per section for myelinated axons
        Returns
        -------
        footprints   : dict
            Dictionnary composed of axon footprint dictionary, the keys are int value
            of the corresponding axon ID
        """
        footprints = {}
        mp_computation = False
        if is_FEM_extra_stim(self.extra_stim):
            mp_computation = (
                self.extra_stim.fenics and self.extra_stim.model.is_multi_proc
            )

            if MCH.do_master_only_work() or mp_computation:
                self.extra_stim.run_model()

        if MCH.do_master_only_work() or mp_computation:
            self.set_axons_parameters(**kwargs)
            for k in range(len(self.axons_diameter)):
                if self.axons_type[k] == 0:
                    axon = unmyelinated(
                        self.axons_y[k],
                        self.axons_z[k],
                        round(self.axons_diameter[k], 2),
                        self.L,
                        ID=k,
                        **self.unmyelinated_param,
                    )
                else:
                    axon = myelinated(
                        self.axons_y[k],
                        self.axons_z[k],
                        round(self.axons_diameter[k], 2),
                        self.L,
                        ID=k,
                        **self.myelinated_param,
                    )
                axon.attach_extracellular_stimulation(self.extra_stim)
                footprints[k] = axon.get_electrodes_footprints_on_axon(
                    save_ftp_only=save_ftp_only, filename=filename
                )
                del axon
            if save_ftp_only:
                json_dump(footprints, filename)
        else:
            pass
        synchronize_processes()
        if not mp_computation:
            self.footprints = MCH.master_broadcasts_array_to_all(footprints)
        else:
            self.footprints = footprints
        self.is_footprinted = True
        return footprints

    def get_electrodes_footprints_on_axons(
        self, save_ftp_only=False, filename="electrodes_footprint.ftpt", **kwargs
    ):
        rise_warning(
            "Deprecation method get_electrodes_footprints_on_axons",
            "\nuse get_electrodes_footprints_on_axons instead",
        )
        self.compute_electrodes_footprints(
            save_ftp_only=save_ftp_only, filename=filename, **kwargs
        )

    def __sim_axon(
        self,
        k,
        is_master_slave=False,
        **kwargs,
    ):
        """
        Internal use only simumlated one axon of the fascicle

        Parameters
        ----------
        k       : int
            ID of the axon to simulate
        """
        ## test axon axons_type[k]
        assert self.axons_type[k] in [0, 1]
        if self.axons_type[k] == 0:
            axon = unmyelinated(
                self.axons_y[k],
                self.axons_z[k],
                round(self.axons_diameter[k], 2),
                self.L,
                ID=k,
                **self.unmyelinated_param,
            )
        else:
            axon = myelinated(
                self.axons_y[k],
                self.axons_z[k],
                round(self.axons_diameter[k], 2),
                self.L,
                ID=k,
                **self.myelinated_param,
            )
        ## add extracellular stimulation
        axon.attach_extracellular_stimulation(self.extra_stim)
        ## add recording mechanism
        if self.record:
            axon.attach_extracellular_recorder(self.recorder)
        # add intracellular stim
        if self.N_intra > 0:
            for j in range(self.N_intra):
                if is_iterable(self.intra_stim_ON[j]):
                    # in this case, the stimulation is possibly not for all axons
                    if self.intra_stim_ON[j][k]:
                        # then stimulation should apply, look for the parameters
                        # get position
                        if is_iterable(self.intra_stim_position[j]):
                            position = self.intra_stim_position[j][k]
                        else:
                            position = self.intra_stim_position[j]
                        # get t_start
                        if is_iterable(self.intra_stim_t_start[j]):
                            t_start = self.intra_stim_t_start[j][k]
                        else:
                            t_start = self.intra_stim_t_start[j]
                        # get duration
                        if is_iterable(self.intra_stim_duration[j]):
                            duration = self.intra_stim_duration[j][k]
                        else:
                            duration = self.intra_stim_duration[j]
                        # get amplitude
                        if is_iterable(self.intra_stim_amplitude[j]):
                            amplitude = self.intra_stim_amplitude[j][k]
                        else:
                            amplitude = self.intra_stim_amplitude[j]
                        # APPLY INTRA CELLULAR STIMULATION
                        axon.insert_I_Clamp(position, t_start, duration, amplitude)
                else:
                    # in this case , stimulate all axons
                    if is_iterable(self.intra_stim_position[j]):
                        position = self.intra_stim_position[j][k]
                    else:
                        position = self.intra_stim_position[j]
                    # get t_start
                    if is_iterable(self.intra_stim_t_start[j]):
                        t_start = self.intra_stim_t_start[j][k]
                    else:
                        t_start = self.intra_stim_t_start[j]
                    # get duration
                    if is_iterable(self.intra_stim_duration[j]):
                        duration = self.intra_stim_duration[j][k]
                    else:
                        duration = self.intra_stim_duration[j]
                    # get amplitude
                    if is_iterable(self.intra_stim_amplitude[j]):
                        amplitude = self.intra_stim_amplitude[j][k]
                    else:
                        amplitude = self.intra_stim_amplitude[j]
                    # APPLY INTRA CELLULAR STIMULATION
                    axon.insert_I_Clamp(position, t_start, duration, amplitude)
        if is_master_slave:
            ## perform simulation
            # loaded_footprints is suppose to be false
            # Request to server instantiate in FEM_stimulation.compute_electrodes_footprints
            # (called by axon.get_electrodes_footprints_on_axon)
            axon.get_electrodes_footprints_on_axon()
            axon_ftpt = None
        else:
            if self.loaded_footprints == True and self.is_footprinted:
                axon_ftpt = True
                if k in self.footprints:
                    axon.footprints = self.footprints[k]
                elif str(k) in self.footprints:
                    axon.footprints = self.footprints[str(k)]
                else:
                    rise_warning("footprints not computed for axon " + str(k))
                    axon_ftpt = False
            else:
                axon_ftpt = False

        axon_sim = axon.simulate(
            t_sim=self.t_sim,
            record_V_mem=self.record_V_mem,
            record_I_mem=self.record_I_mem,
            record_I_ions=self.record_I_ions,
            record_g_mem=self.record_g_mem,
            record_g_ions=self.record_g_ions,
            record_particles=self.record_particles,
            loaded_footprints=axon_ftpt,
        )
        del axon
        return axon_sim

    def simulate(
        self,
        PostProc_Filtering=None,
        save_V_mem = False,
        **kwargs,
    ):
        """
        Simulates the fascicle using neuron framework. Parallel computing friendly. Does not return
        results (possibly too large in memory and complex with parallel computing), but instead
        creates a folder and store fascicle configuration and all axons results.
        On the fly post processing is possible by specifying an additional script.

        Parameters
        ----------
        t_sim               : float
            total simulation time (ms), by default 20 ms
        record_V_mem        : bool
            if true, the membrane voltage is recorded, set to True by default
        record_I_mem        : bool
            if true, the membrane current is recorded, set to False by default
        record_I_ions       : bool
            if true, the ionic currents are recorded, set to False by default
        record_particules   : bool
            if true, the marticule states are recorded, set to False by default
        loaded_footprints          : dict or bool
            Dictionnary composed of axon footprint dictionary, the keys are int value
            of the corresponding axon ID. if type is bool, fascicle footprints attribute is used
            if None, footprins calculated during the simulation, by default None
        save_V_mem          : bool
            if true, all membrane voltages values are stored in results whe basic postprocessing is
            applied. Can be heavy ! False by default
        save_path           : str
            name of the folder to store results of the fascicle simulation.
        Unmyelinated_model  : str
            model for unmyelinated fibers, by default 'Rattay_Aberham'
        Myelinated_model    : str
            model for myelinated fibers, by default 'MRG'
        myelinated_seg_per_sec        : int
            number of segment per section for myelinated axons
        PostProc_Filtering  : float, list, array, np.array
            value or iterable values for basic post proc filtering. If None specified, no filtering
            is performed
        postproc_script     : str
            path to a postprocessing file. If specified, the basic post processing is not
            performed, and all postprocessing have to be handled by user. The specified script
            can access global and local variables. Can also be key word ('Vmem_plot' or 'scarter')
            to use script saved in OTF_PP folder, use with caution
        return_parameters_only        : bool
            if True the results of axon simulations are integrated into fasc_sim return after simulation
            use with caution: can increase a lot computational memory
        save_results                  : bool
            if False disable the result storage
            can only be False if return_parameters_only is False

        Return
        ------
            fasc_sim    : sim_results
                results of the simulation
        """
        if self.postproc_script is not None:
            import nrv  # not ideal at all but gives direct acces to nrv in the postprocessing script
        fasc_sim = super().simulate(**kwargs)
        if not MCH.do_master_only_work():
            fasc_sim = sim_results({})
        # disable the result storage only if results are fully return
        # To use with caution (mostly for optimisatino)
        if not self.save_results:
            if self.return_parameters_only:
                rise_warning(
                    "Fascicle's simulation parameters are misused",
                    "results are neither saved or return", 
                    "save anyway")
                self.save_results = True

        if len(self.NoR_relative_position) == 0:
            self.generate_random_NoR_position()
        # create folder and save fascicle config
        folder_name = self.save_path + "Fascicle_" + str(self.ID)
        if MCH.do_master_only_work() and self.save_results:
            create_folder(folder_name)
            self.config_filename = folder_name + "/00_Fascicle_config.json"
            fasc_sim.save(save=True, fname=self.config_filename)
        else:
            pass
        # impose myelinated_nseg_per_sec if footprint are loaded
        self.loaded_footprints = self.loaded_footprints and self.is_footprinted
        if self.loaded_footprints:
            kwargs["myelinated_nseg_per_sec"] = self.myelinated_param["Nseg_per_sec"]
            kwargs["unmyelinated_nseg"] = self.unmyelinated_param["Nseg_per_sec"]
        self.set_axons_parameters(**kwargs)
        self.processes_status = ["computing" for _ in range(MCH.size)]
        # Check if FEM should be done in parallel and do it if required
        mp_computation = False
        if is_FEM_extra_stim(self.extra_stim):
            mp_computation = (
                self.extra_stim.fenics and self.extra_stim.model.is_multi_proc
            )
        if mp_computation and not self.loaded_footprints:
            self.compute_electrodes_footprints()
            self.loaded_footprints = True
        # create ID for all axons
        axons_ID = np.arange(len(self.axons_diameter))

        # FEM STIMULATION IN PARALLEL: master computes FEM (only one COMSOL licence, other computes axons)
        if (
            self.extra_stim is not None
            and self.loaded_footprints == False
            and not is_analytical_extra_stim(self.extra_stim)
            and not MCH.is_alone()
            and not mp_computation
        ):
            is_master_slave = True
            # master solves FEM model
            if MCH.do_master_only_work():
                self.extra_stim.run_model()
            else:
                if self.extra_stim.fenics:
                    self.extra_stim.run_model()
            # synchronize all process
            synchronize_processes()
            # split the job
            this_core_mask = MCH.split_job_from_arrays_to_slaves(
                len(self.axons_diameter)
            )
            ## Axons simulation handle with Master/Slaves configuration:
            # Master: acts as a server, accepting request to compute external potential
            #   Also, update status of slaves when simulations complete
            # Slaves: computes neurons sending requests to the master
            #   (see FEM_stimulation.compute_electrodes_footprints)
            #   When simulations completed or interupted by an error,
            #   send status update ('success' or 'error') to Master
            ## MASTER ##
            if MCH.do_master_only_work():
                self.processes_status = ["server"] + [
                    "computing" for _ in range(MCH.size - 1)
                ]
                while "computing" in self.processes_status:
                    data = MCH.recieve_data_from_slave()
                    if "status" in data:
                        self.processes_status[data["rank"]] = data["status"]
                    else:
                        this_core_mask[data["ID"]] = True
                        V = self.extra_stim.model.get_potentials(
                            data["x"], data["y"], data["z"]
                        )
                        back_data = {"V": V}
                        MCH.send_back_array_to_dest(back_data, data["rank"])
                if "error" in self.processes_status:
                    errors_rank = [
                        i for i, x in enumerate(self.processes_status) if x == "error"
                    ]
                    errors_ax = [i for i, x in enumerate(this_core_mask) if x == False]
                    if len(errors_ax) > 0:
                        rise_warning(
                            "an issue occured during the simulation in rank: ",
                            errors_rank,
                            "\nThe following axons could not be computed: ",
                            errors_ax,
                        )
                    self.processes_status[data["rank"]]
                    self.processes_status[0] = ["error"]
                else:
                    self.processes_status[0] = ["success"]
            else:
                ## SLAVES ##
                try:
                    for k in this_core_mask:
                        axon_sim = self.__sim_axon(k,is_master_slave)
                        # postprocessing and data reduction
                        # (cannot be added to __sim.axon beceause nrv would not be global anymore)
                        if self.postproc_script.lower() in OTF_PP_library:
                            self.postproc_script = OTF_PP_path + self.postproc_script.lower()
                        elif self.postproc_script.lower() + ".py" in OTF_PP_library:
                            self.postproc_script = (
                                OTF_PP_path + self.postproc_script.lower() + ".py"
                            )
                        with open(self.postproc_script) as f:
                            code = compile(f.read(), self.postproc_script, "exec")
                            exec(code, globals(), locals())
                        if self.save_results:
                            ## store results
                            ax_fname = "sim_axon_" + str(k) + ".json"
                            #save_axon_results_as_json(axon_sim, folder_name + "/" + ax_fname)
                            axon_sim.save(save=True, fname=folder_name + "/" + ax_fname)
                        if not self.return_parameters_only:
                            fasc_sim.update({"axon"+str(k):axon_sim})
                    data = {"rank": MCH.rank, "status": "success"}
                    MCH.send_data_to_master(data)
                except KeyboardInterrupt:
                    raise KeyboardInterrupt
                except:
                    data = {"rank": MCH.rank, "status": "error"}
                    MCH.send_data_to_master(data)
                    rise_error(
                        traceback.format_exc()
                    )
                # sum up all recorded extracellular potential if applicable
            synchronize_processes()
            if self.record:
                self.recorder.gather_all_recordings(master_computed=False)
        ###### NO STIM OR ANALYTICAL STIM: all in parallel, OR FEM STIM NO PARALLEL
        else:
            is_master_slave = False
            ## split the job between Cores/Computation nodes
            this_core_mask = MCH.split_job_from_arrays(len(self.axons_diameter))
            ## perform simulations
            if MCH.is_alone() and self.verbose:
                pass_info("Simulating axons in fascicle " + str(self.ID))
            for k in this_core_mask:
                if MCH.is_alone() and self.verbose:
                    pass_info("\t Axon " + f"{k+1}" + "/" + str(self.N), end="\r")
                axon_sim = self.__sim_axon(k, is_master_slave)
                # postprocessing and data reduction
                # (cannot be added to __sim.axon beceause nrv would not be global anymore)
                if self.postproc_script.lower() in OTF_PP_library:
                    self.postproc_script = OTF_PP_path + self.postproc_script.lower()
                elif self.postproc_script.lower() + ".py" in OTF_PP_library:
                    self.postproc_script = OTF_PP_path + self.postproc_script.lower() + ".py"
                with open(self.postproc_script) as f:
                    code = compile(f.read(), self.postproc_script, "exec")
                    exec(code, globals(), locals())
                ## store results
                if self.save_results:
                    ax_fname = "sim_axon_" + str(k) + ".json"
                    #save_axon_results_as_json(axon_sim, folder_name + "/" + ax_fname)
                    axon_sim.save(save=True, fname=folder_name + "/" + ax_fname)
                if not self.return_parameters_only:
                    fasc_sim.update({"axon"+str(k):axon_sim})
            # sum up all recorded extracellular potential if applicable
            synchronize_processes()
            if self.record:
                self.recorder.gather_all_recordings()
        if not self.return_parameters_only:
            synchronize_processes()
            fasc_sim = MCH.gather_jobs(fasc_sim)
        if MCH.is_alone() and self.verbose:
            pass_info("... Simulation done")
            fasc_sim["is_simulated"] = True
        self.is_simulated = True
        return fasc_sim



from ..backend.NRV_Class import NRV_class, load_any
from ..backend.NRV_Simulable import NRV_simulable
from ..backend.MCore import MCH, synchronize_processes

class CostFunction(NRV_class):
    """
    A class to define cost from position input vector

    Parameters
    ----------
    context_modifier   : funct
        function creating a context from a particle position, parameters:
            position        : particle position in each dimensions (array)
            t_sim           : simulation time (float)
            dt              : time step of the context (float)
            kwargs          : keys word arguments
        returns
            context        : values of a context (array)
    simulate_context   : funct
        function which perform the NEURON simulation, parameters:
            context        : context to use for the simulation (array)
            t_sim           : simulation time (float)
            dt              : time step of the context (float)
            kwargs          : keys word arguments
        returns
            results        : axon simulaiton results (dictionary)
    cost_evaluation            : funct
        function calculating a cost from a axon simulation results, parameters:
            results        : axon simulaiton results (dictionary)
            kwargs          : keys word arguments
        returns
            cost              : cost evaluation value (float)
    kwargs_CM           : dict
        key word arguments of context_modifier function, by default {}
    kwargs_S           : dict
        key word arguments of simulate_context function, by default {}
    kwargs_CE            : dict
        key word arguments of cost_evaluation function, by default {}
    t_sim           : float
        time of the simulation on wich the cost will be calculated in ms (ms), by default 100ms
    dt              : float
        simulation time stem for neuron (ms), by default 1us
    part_filter         : funct or None
        function wich return a filtered postition from a position, if None, position is unfiltered,
        by default None
    saver         : funct or None
        function added at the end after calculating the cost for personalized savings, if None,
        nothing will be saved,  by default None

    """

    def __init__(
        self,
        static_context,
        context_modifier,
        cost_evaluation,
        simulation=None,
        kwargs_CM={},
        kwargs_S={},
        kwargs_CE={},
        filter=None,
        saver=None,
        file_name="cost_saver.csv",
        **kawrgs
    ):
        super().__init__()
        self.static_context = static_context
        self.context_modifier = context_modifier
        self.simulation = simulation
        self.cost_evaluation = cost_evaluation
        self.kwargs_CM = kwargs_CM
        self.kwargs_S = kwargs_S
        self.kwargs_CE = kwargs_CE
        self.filter = filter
        self.saver = saver
        self.file_name = file_name

        self.kwargs_CM.update(kawrgs)
        self.kwargs_S.update(kawrgs)
        self.kwargs_CE.update(kawrgs)

        self.simulation_context = None
        self.results = None
        self.cost = None

        self._MCore_CostFunction = False
        self.__check_cost_function()

    def __clear_results(self):
        del self.simulation_context
        del self.results
        self.simulation_context = None
        self.results = None

    def __check_cost_function(self):
        static_contect = load_any(self.static_context)
        if (
            not MCH.is_alone() and
            static_contect.nrv_type in ["fascicle","nerve"]
        ):
            self._MCore_CostFunction = True
        del static_contect

    def simulate_context(self):
        if callable(self.simulation):
            results = self.simulation(self.simulation_context, **self.kwargs_S)
        if self.simulation is None:
            if "return_parameters_only" not in self.kwargs_S:
                self.kwargs_S["return_parameters_only"] = False
            results = self.simulation_context(**self.kwargs_S)
        return results

    def __call__(self, X):
        # Broadcasting to slave wich are waiting for simulation
        if MCH.do_master_only_work() and self._MCore_CostFunction:
            slave_status = {"status": "Simulate", "X":X}
            MCH.master_broadcasts_array_to_all(slave_status)
        if self._MCore_CostFunction:
            synchronize_processes()

        # Filter
        if self.filter is not None:
            X_ = self.filter(X)
        else:
            X_ = X

        # Interpolation
        self.simulation_context = self.context_modifier(X_, self.static_context, **self.kwargs_CM)

        # Simulation
        self.results = self.simulate_context()
        if MCH.do_master_only_work() or not self._MCore_CostFunction:
            # Cost calculation
            self.cost = self.cost_evaluation(self.results, **self.kwargs_CE)
            # Savings
            if self.saver is not None:
                data = {
                    "position": X,
                    "context": self.simulation_context,
                    "results": self.results,
                    "cost": self.cost,
                }
                self.saver(data, file_name=self.file_name)
            self.__clear_results()
            return self.cost
        else:
            return 0



"""
NRV-extracellular contexts
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler

import numpy as np

from ..backend.file_handler import json_load
from ..backend.log_interface import rise_error, rise_warning
from ..backend.MCore import MCH
from ..backend.NRV_Class import NRV_class
from .electrodes import is_analytical_electrode, is_FEM_electrode, check_electrodes_overlap
from .FEM.COMSOL_model import COMSOL_model, COMSOL_Status
from .FEM.FENICS_model import FENICS_model
from .materials import load_material, is_mat
from .stimulus import get_equal_timing_copies

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()


def is_extra_stim(test_stim):
    """
    check if an object is a stimulation, return True if yes, else False
    """
    Flag = (
        isinstance(test_stim, stimulation)
        or isinstance(test_stim, FEM_stimulation)
        or isinstance(test_stim, extracellular_context)
    )
    return Flag


def is_analytical_extra_stim(test_stim):
    """
    check if an object is a stimulation (analytical only), return True if yes, else False
    """
    return isinstance(test_stim, stimulation)


def is_FEM_extra_stim(test_stim):
    """
    check if an object is a FEM stimulation, return True if yes, else False
    """
    return isinstance(test_stim, FEM_stimulation)


def load_any_extracel_context(data):
    """
    return any kind of extracellular context properties from a dictionary or a json file

    Parameters
    ----------
    data    : str or dict
        json file path or dictionary containing extracel_context information
    """
    if type(data) == str:
        context_dic = json_load(data)
    else:
        context_dic = data

    if context_dic["type"] is None:
        extracel = extracellular_context()
    elif context_dic["type"] == "stimulation":
        extracel = stimulation()
    elif context_dic["type"] == "FEM_stim":
        extracel = FEM_stimulation(data["model_fname"], comsol=False)
    else:
        rise_error("extra cellular context type not recognizede")

    extracel.load(context_dic)
    return extracel


class extracellular_context(NRV_class):
    """
    extracellular_context is a class to handle the computation of the extracellular voltage field induced by the electrical stimulation.
    This class should not be used directly by user, but user friendly classes (for Analitycal or FEM based computations) inherits from extracellular_context.
    """

    def __init__(self):
        """
        Instrantiation an extracellular_context object, empty shell to store electrodes and stimuli
        """
        super().__init__()
        self.type = "extracellular_context"
        # empty list to store electrodes and corresponding stimuli
        self.electrodes = []
        self.stimuli = []
        # list for synchronised stimuli
        self.synchronised = False
        self.synchronised_stimuli = []
        self.global_time_serie = []
        self.type = None

    def save_extracel_context(self, save=False, fname="extracel_context.json"):
        rise_warning("save_extracel_context is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_extracel_context(self, data="extracel_context.json"):
        rise_warning("load_extracel_context is a deprecated method use load")
        self.load(data=data)

    ##
    def is_empty(self):
        """
        check if a stimulation object is empty (No electrodes and stimuli, no external field can be computed).
        Returns True if empty, else False.

        Returns
        -------
        bool
            True if a simulation is empty, else False
        """
        return self.electrodes == []

    def translate(self, x=None, y=None, z=None):
        """
        Move extracellular context electrodes by group translation

        Parameters
        ----------
        x   : float
            x axis value for the translation in um
        y   : float
            y axis value for the translation in um
        z   : float
            z axis value for the translation in um
        """
        for elec in self.electrodes:
            elec.translate(x=x, y=y, z=z)

    def add_electrode(self, electrode, stimulus):
        """
        Add a stimulation electrode and its stimulus to the stimulation.

        Parameters
        ----------
        electrode   : electrode object
            see Electrode.py or electrode object help for further details
        stimulus    : stimulus object
            see Stimulus.py or stimulus object help for further details
        """
        if self.electrodes == []:
            self.electrodes.append(electrode)
            self.stimuli.append(stimulus)
        else:
            electrode.set_ID_number(self.electrodes[-1].get_ID_number() + 1)
            self.electrodes.append(electrode)
            self.stimuli.append(stimulus)
        self.synchronised = False

    def reset_stimuli(self):
        self.stimuli = []
        self.synchronised_stimuli = []
        self.synchronised = False
        self.global_time_serie = []

    def reset_electrodes(self):
        self.electrodes = []
        self.reset_stimuli()

    def synchronise_stimuli(self, snap_time=False, dt_min=0.001):
        """
        Synchronise all stimuli before simulation. Copies of the stimuli are created with the global number of samples
        from merging all stimuli time samples. Original stimuli are not affected.
        """
        if not (self.synchronised or self.is_empty()):
            if len(self.electrodes) == 1:
                self.synchronised_stimuli.append(self.stimuli[0])
            elif len(self.electrodes) == 2:
                stim_a, stim_b = get_equal_timing_copies(
                    self.stimuli[0], self.stimuli[1]
                )
                self.synchronised_stimuli.append(stim_a)
                self.synchronised_stimuli.append(stim_b)
            else:
                # init : put first two and synchronize them
                stim_a, stim_b = get_equal_timing_copies(
                    self.stimuli[0], self.stimuli[1]
                )
                self.synchronised_stimuli.append(stim_a)
                self.synchronised_stimuli.append(stim_b)
                # remaining stimuli to handle
                unsynchronized_stim = self.stimuli[2:]
                for stimulus in unsynchronized_stim:
                    # synchronise all the previously synchronised with the pending one
                    for s in self.synchronised_stimuli:
                        s.insert_samples(stimulus.t)
                    # synchronise the pending one with the already synchronised and add it to synchronised stim
                    stimulus.insert_samples(self.synchronised_stimuli[0].t)
                    self.synchronised_stimuli.append(stimulus)
            # if snap_time is true synchronised_stimuli overlaping points are removed
            if snap_time:
                for stim in self.synchronised_stimuli:
                    stim.snap_time(dt_min)
            # anyway, take the first stimulus time serie as the global one
            self.global_time_serie = self.synchronised_stimuli[0].t
        self.synchronised = True

    def compute_vext(self, time_index):
        """
        Compute the external potential on a array of coordinate for a time sample of all synchronised stimuli with all
        electrodes.

        Parameters
        ----------
        time_index  : int
            time index of the synchronised stimuli to compute the field at.
            NB: as a safeguard, if the time_index is out of the sample list a null potential is returned

        Returns
        -------
        Vext : np.array
            external potential for the specified positions, in mV (numpy array)
        """
        Vext = np.zeros(len(self.electrodes[0].footprint))
        if not self.synchronised:
            self.synchronise_stimuli()
        if time_index < len(
            self.global_time_serie
        ):  # requested time index is in stimulus range (safeguard)
            for k in range(len(self.electrodes)):
                Istim = self.synchronised_stimuli[k].s[time_index]
                vext_elec = self.electrodes[k].compute_field(Istim)
                Vext = Vext + vext_elec
        return Vext

    def set_electrodes_footprints(self, footprints):
        """
        set the footprints for all electrodes from existing array

        Parameters
        ----------
        footprints  : list of array like
            list footprint for each electode of the extracellular context
        """
        i = 0
        if len(footprints) == len(self.electrodes):
            for i, electrode in enumerate(self.electrodes):
                if i in footprints:
                    electrode.set_footprint(footprints[i])
                else:
                    electrode.set_footprint(footprints[str(i)])
        else:
            rise_error("Footprint number different than electrode number")

    def clear_electrodes_footprints(self):
        """
        clear the footprints for all electrodes from existing array
        """
        for electrode in self.electrodes:
            electrode.clear_footprint()

    def change_stimulus_from_elecrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        Parameters:
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                New stimulus to set
        """
        if ID_elec < len(self.stimuli):
            self.stimuli[ID_elec] = stimulus
            self.synchronised_stimuli = []
            self.synchronised = False
        else:
            rise_warning(
                "Only",
                len(self.stimuli),
                "electrode in extracellular_context:",
                ID_elec,
                "is not too big",
            )


class stimulation(extracellular_context):
    """
    Stimulation object are designed to connect all other objects requierd to analyticaly compute the external potential voltage for axons :
    - the material surrounding the axon (only one)
    - a list of electrode(s)
    - a list of corresponding current stimuli
    This class inherits from extracellular_context.
    """

    def __init__(self, material="endoneurium_ranck"):
        """
        Implement a stimulation object.

        Parameters
        ----------
        material    : str or material object
            extracellular medium see Material.py or material object help for further details
        """
        super().__init__()
        if is_mat(material):
            self.material = material
        else:
            self.material = load_material(material)
        self.type = "stimulation"

    def add_electrode(self, electrode, stimulus):
        """
        Add a stimulation electrode and its stimulus to the stimulation, only if the electrode is analytically described.

        Parameters
        ----------
        electrode   : electrode object
            see Electrode.py or electrode object help for further details
        stimulus    : stimulus object
            see Stimulus.py or stimulus object help for further details
        """
        if is_analytical_electrode(electrode):
            if self.electrodes == []:
                self.electrodes.append(electrode)
                self.stimuli.append(stimulus)
            else:
                electrode.set_ID_number(self.electrodes[-1].get_ID_number() + 1)
                self.electrodes.append(electrode)
                self.stimuli.append(stimulus)
            self.synchronised = False

    def compute_electrodes_footprints(self, x, y, z, ID=0):
        """
        Compute the footprints for all electrodes

        Parameters
        ----------
        x           : np.array
            x position at which to compute the field, in um
        y           : float
            y position at which to compute the field, in um
        z           : float
            z position at which to compute the field, in um
        ID          : int
            axon ID, unused here, added to fit FEM_stimulation declaration of same method
        """
        for electrode in self.electrodes:
            electrode.compute_footprint(
                np.asarray(x), np.asarray(y), np.asarray(z), self.material
            )


class FEM_stimulation(extracellular_context):
    """
    FEM_based_simulation object are designed to connect all other objects requierd to compute the external potential voltage for axons using FEM :
    - the materials for the FEM stimulation : endoneurium, perineurium, epineurium and external material
    - a list of electrode(s)
    - a list of corresponding current stimuli
    """

    def __init__(
        self,
        model_fname=None,
        endo_mat="endoneurium_ranck",
        peri_mat="perineurium",
        epi_mat="epineurium",
        ext_mat="saline",
        comsol=True,
        Ncore=None,
    ):
        """
        Implement a FEM_based_stimulation object.

        Parameters
        ----------
        simuluation_fname   : str
            name of the simluation file to compute the field
        endo_mat            : material object
            specification of the endoneurium material, see Material.py or material object help for further details
        peri_mat            : material object
            specification of the perineurium material, see Material.py or material object help for further details
        epi_mat             : material object
            specification of the epineurium material, see Material.py or material object help for further details
        ext_mat             : material object
            specification of the external material (everything but the nerve), see Material.py or material object help for further details
        """
        super().__init__()
        self.electrodes_label = []
        self.model_fname = model_fname
        self.setup = False
        self.is_run = False
        self.Ncore = Ncore
        ## get material properties and add to model

        if is_mat(endo_mat):
            self.endoneurium = endo_mat
            self.endo_mat_file = None
        else:
            self.endoneurium = load_material(endo_mat)
            self.endo_mat_file = endo_mat
        if is_mat(peri_mat):
            self.perineurium = peri_mat
            self.peri_mat_file = None
        else:
            self.perineurium = load_material(peri_mat)
            self.peri_mat_file = peri_mat
        if is_mat(epi_mat):
            self.epineurium = epi_mat
            self.epi_mat_file = None
        else:
            self.epineurium = load_material(epi_mat)
            self.epi_mat_file = epi_mat
        if is_mat(ext_mat):
            self.external_material = ext_mat
            self.ext_mat_file = None
        else:
            self.external_material = load_material(ext_mat)
            self.ext_mat_file = ext_mat

        self.type = "FEM_stim"

        self.comsol = comsol and not (model_fname is None)
        self.fenics = not self.comsol
        ## load model
        if self.comsol:
            if MCH.do_master_only_work():
                self.model = COMSOL_model(self.model_fname, Ncore=self.Ncore)
            else:
                # check that COMSOL is not turned OFF in order to continue
                if not COMSOL_Status:
                    rise_warning(
                        "Slave process abort as axon is supposed to used a COMSOL FEM and COMSOL turned off",
                        abort=True,
                    )
        else:
            self.model = FENICS_model(Ncore=self.Ncore)

    def __del__(self):
        if (
            MCH.do_master_only_work() and COMSOL_Status and self.comsol
        ):  # added for safe del in case of COMSOL status turned OFF
            self.model.close()
            del self.model

    def set_Ncore(self, N):
        """
        Set the number of cores to use for the FEM

        Parameters
        ----------
        N       : int
            Number of cores to set
        """
        self.model.set_Ncore(N=N)

    ## Save and Load mehtods

    def save(self, save=False, fname="extracel_context.json", blacklist=[], **kwargs):
        """
        Return extracellular context as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "extracel_context.json"

        Returns
        -------
        context_dic : dict
            dictionary containing all information
        """
        if self.comsol:
            blacklist += ["model"]

        return super().save(save=save, fname=fname, blacklist=blacklist, **kwargs)

    def load(self, data, C_model=False, **kwargs):
        """
        Load all extracellular context properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing extracel_context information
        """
        super().load(data, **kwargs)

        if C_model:
            if MCH.do_master_only_work():
                self.model = COMSOL_model(self.model_fname)
            else:
                # check that COMSOL is not turned OFF in order to continue
                if not COMSOL_Status:
                    rise_warning(
                        "Slave process abort as axon is supposed to used a COMSOL FEM and COMSOL turned off",
                        abort=True,
                    )

    def reshape_outerBox(self, Outer_D, res="default"):
        """
        Reshape the size of the FEM simulation outer box

        Parameters
        ----------
        outer_D : float
            FEM simulation outer box diameter, in mm, WARNING, this is the only parameter in mm !
        res         : float or "default"
            mesh resolution for fenics_model cf NerveMshCreator, use with caution, by default "default"
        """
        if MCH.do_master_only_work():
            if self.comsol:
                self.model.set_parameter("Outer_D", str(Outer_D) + "[mm]")
            else:
                self.model.reshape_outerBox(Outer_D, res=res)

    def reshape_nerve(self, Nerve_D, Length, y_c=0, z_c=0, res="default"):
        """
        Reshape the nerve of the FEM simulation

        Parameters
        ----------
        Nerve_D                 : float
            Nerve diameter, in um
        Length                  : float
            Nerve length, in um
        y_c                     : float
            Nerve center y-coordinate in um, 0 by default
        z_c                     : float
            Nerve z-coordinate center in um, 0 by default
        res         : float or "default"
            mesh resolution for fenics_model cf NerveMshCreator, use with caution, by default "default"
        """
        if MCH.do_master_only_work():
            if self.comsol:
                self.model.set_parameter("Nerve_D", str(Nerve_D) + "[um]")
                self.model.set_parameter("Length", str(Length) + "[um]")
                self.model.set_parameter("Nerve_y_c", str(y_c) + "[um]")
                self.model.set_parameter("Nerve_z_c", str(z_c) + "[um]")
            else:
                self.model.reshape_nerve(
                    Nerve_D=Nerve_D, Length=Length, y_c=y_c, z_c=z_c, res=res
                )

    def reshape_fascicle(
        self, Fascicle_D, y_c=0, z_c=0, ID=None, Perineurium_thickness=5, res="default"
    ):
        """
        Reshape a fascicle of the FEM simulation

        Parameters
        ----------
        Fascicle_D  : float
            Fascicle diameter, in um
        y_c         : float
            Fascicle center y-coodinate in um, 0 by default
        z_c         : float
            Fascicle center y-coodinate in um, 0 by default
        Perineurium_thickness   :float
            Thickness of the Perineurium sheet surounding the fascicles in um, 5 by default
        ID          : int
            If the simulation contains more than one fascicles, ID number of the fascicle to reshape as in COMSOL
        res         : float or "default"
            mesh resolution for fenics_model cf NerveMshCreator, use with caution, by default "default"
        """
        if MCH.do_master_only_work():
            if self.comsol:
                if ID is None:
                    self.model.set_parameter("Fascicle_D", str(Fascicle_D) + "[um]")
                    self.model.set_parameter("Fascicle_y_c", str(y_c) + "[um]")
                    self.model.set_parameter("Fascicle_z_c", str(z_c) + "[um]")
                else:
                    self.model.set_parameter(
                        "Fascicle_" + str(ID) + "_D", str(Fascicle_D) + "[um]"
                    )
                    self.model.set_parameter(
                        "Fascicle_" + str(ID) + "_y_c", str(y_c) + "[um]"
                    )
                    self.model.set_parameter(
                        "Fascicle_" + str(ID) + "_z_c", str(z_c) + "[um]"
                    )
                self.model.set_parameter(
                    "Perineurium_thickness", str(Perineurium_thickness) + "[um]"
                )
            else:
                self.model.reshape_fascicle(
                    Fascicle_D=Fascicle_D,
                    y_c=y_c,
                    z_c=z_c,
                    ID=ID,
                    Perineurium_thickness=Perineurium_thickness,
                    res=res,
                )

    def remove_fascicles(self, ID=None):
        """
        remove a fascicle of the FEM simulation

        Parameters
        ----------
        ID          : int, None
            ID number of the fascicle to remove, if None, remove all fascicles, by default None
        """
        if MCH.do_master_only_work():
            self.model.remove_fascicles(ID=ID)

    def add_electrode(self, electrode, stimulus):
        """
        Add a stimulation electrode and its stimulus to the stimulation, only it the electrode is FEM based.

        Parameters
        ----------
        electrode   : electrode object
            see Electrode.py or electrode object help for further details
        stimulus    : stimulus object or list[stimulus]
            see Stimulus.py or stimulus object help for further details, for Multipolar electrode:
            if stimulus a list of situmulus one stimulus set for each active site
            else
        """
        is_overlaping = False
        for elec in self.electrodes:
            is_overlaping = is_overlaping or check_electrodes_overlap(elec, electrode)

        if is_overlaping:
            rise_warning("overlaping electrod: not added to context")
        else:
            if is_FEM_electrode(electrode):
                if not electrode.is_multipolar:
                    if not self.electrodes == []:
                        electrode.set_ID_number(self.electrodes[-1].get_ID_number() + 1)
                    self.electrodes.append(electrode)
                    self.electrodes_label.append(electrode.label)
                    self.stimuli.append(stimulus)
                else:
                    if np.iterable(stimulus):
                        stimuli = stimulus
                    else:
                        rise_warning(
                            "Only one stimulus for a multipolar electrode, it will be set for all active sites"
                        )
                        stimuli = [stimulus for k in range(electrode.N_contact)]
                    for E in range(electrode.N_contact):
                        if not self.electrodes == []:
                            electrode.set_ID_number(
                                self.electrodes[-1].get_ID_number() + 1
                            )
                        self.electrodes.append(electrode)
                        self.electrodes_label.append(electrode.label + "_" + str())
                        self.stimuli.append(stimuli[E])

                if self.fenics and MCH.do_master_only_work():
                    electrode.parameter_model(self.model)
                self.synchronised = False
                self.setup = False
                self.is_run = False

    def setup_FEM(self):
        """
        Parameter a model with all added electrodes parameters, material parameters, build geometry and mesh
        """
        # parameter electrodes
        if self.comsol:
            for electrode in self.electrodes:
                electrode.parameter_model(self.model)
                # parameter materials
                self.model.set_parameter(
                    "Outer_conductivity", str(self.external_material.sigma) + "[S/m]"
                )
                self.model.set_parameter(
                    "Epineurium_conductivity", str(self.epineurium.sigma) + "[S/m]"
                )
                self.model.set_parameter(
                    "Perineurium_conductivity", str(self.perineurium.sigma) + "[S/m]"
                )
                if self.endoneurium.is_isotropic():
                    self.model.set_parameter(
                        "Endoneurium_conductivity_xx",
                        str(self.endoneurium.sigma) + "[S/m]",
                    )
                    self.model.set_parameter(
                        "Endoneurium_conductivity_yy",
                        str(self.endoneurium.sigma) + "[S/m]",
                    )
                    self.model.set_parameter(
                        "Endoneurium_conductivity_zz",
                        str(self.endoneurium.sigma) + "[S/m]",
                    )
                else:
                    self.model.set_parameter(
                        "Endoneurium_conductivity_xx",
                        str(self.endoneurium.sigma_xx) + "[S/m]",
                    )
                    self.model.set_parameter(
                        "Endoneurium_conductivity_yy",
                        str(self.endoneurium.sigma_yy) + "[S/m]",
                    )
                    self.model.set_parameter(
                        "Endoneurium_conductivity_zz",
                        str(self.endoneurium.sigma_zz) + "[S/m]",
                    )
                ## create geometry and mesh
                self.model.build_and_mesh()
        else:
            self.model.build_and_mesh()
            self.model.set_materials(
                Endoneurium_mat=self.endo_mat_file,
                Outer_mat=self.ext_mat_file,
                Perineurium_mat=self.peri_mat_file,
                Epineurium_mat=self.epi_mat_file,
            )
            self.model.setup_simulations()
        self.setup = True

    def run_model(self):
        """
        Set materials properties, build geometry and mesh if not already done and solve the FEM model all in one.
        """
        if MCH.do_master_only_work() or self.model.is_multi_proc:
            if not self.setup:
                self.setup_FEM()
            if not self.is_run:
                self.model.solve()
                self.is_run = True

    def compute_electrodes_footprints(self, x, y, z, ID):
        """
        Compute the footprints for all electrodes

        Parameters
        ----------
        x           : np.array
            x position at which to compute the field, in um
        y           : float
            y position at which to compute the field, in um
        z           : float
            z position at which to compute the field, in um
        ID          : int
            ID of the axon
        """

        if MCH.do_master_only_work() or (self.fenics and self.model.is_multi_proc):
            # test if the model is not solve and run simulation
            if not self.model.is_computed:
                self.run_model()
            # get all potentials
            V = self.model.get_potentials(x, y, z)
        else:
            # send a request to the master to get all potentials
            data = {"rank": MCH.rank, "x": x, "y": y, "z": z, "ID": ID}
            MCH.send_data_to_master(data)
            # listen to the master to get V
            V = MCH.recieve_potential_array_from_master()["V"]
        if len(self.electrodes) > 1:
            # sort electrodes by alphabetical order (COMSOL files should perform the parametric sweep by alphabetical order)
            sorter = np.argsort(self.electrodes_label)
            self.electrodes_label = list(np.asarray(self.electrodes_label)[sorter])
            self.electrodes = list(np.asarray(self.electrodes)[sorter])
            self.stimuli = list(np.asarray(self.stimuli)[sorter])
            # set the footprints
            for k, electrode in enumerate(self.electrodes):
                self.electrodes[k].set_footprint(V[:, k])
                # electrode.set_footprint(V[:, k])
        else:
            self.electrodes[0].set_footprint(V)



"""
NRV-Nerves
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import numpy as np

from ..backend.log_interface import pass_info, rise_warning
from ..backend.NRV_Class import load_any
from ..backend.NRV_Simulable import NRV_simulable
from ..fmod.stimulus import stimulus
from .fascicles import *


#################
## Nerve class ##
#################
class nerve(NRV_simulable):
    """A nerve in NRV is defined as:
        - a list of fascicles
        - a list of materials
        - a kind of extracellular context (analytical or FEM based)

    a nerve can be instrumented by adding couples of electrodes+stimulus
    """

    def __init__(self, Length=10000, Diameter=100, Outer_D=5, ID=0, **kwargs):
        """
        Instanciates an empty nerve.

        Parameters
        ----------
        dt              : float
            simulation time stem for Neuron (ms), by default 1us
        Nseg_per_sec    : float
            number of segment per section in Neuron. If set to 0, the number of segment per section is calculated with the d-lambda rule
        freq            : float
            frequency of the d-lambda rule (Hz), by default 100Hz
        freq_min        : float
            minimum frequency for the d-lambda rule when meshing is irregular, 0 for regular meshing
        v_init          : float
            initial value for the membrane voltage (mV), specify None for automatic model choice of v_init
        T               : float
            temperature (C), specify None for automatic model choice of temperature
        ID              : int
            axon ID, by default set to 0
        threshold       : int
            membrane voltage threshold for spike detection (mV), by default -40mV
        Adelta_limit    : float
            limit diameter between A-delta models (thin myelinated) and myelinated models for axons
        """
        super().__init__()

        # to add to a fascicle/nerve common mother class
        self.save_path=""
        self.verbose=False
        self.postproc_script="default"
        self.return_parameters_only = True
        self.loaded_footprints = True

        self.type = "nerve"
        self.L = Length
        self.D = Diameter
        self.y_grav_center = 0
        self.z_grav_center = 0
        self.Outer_D = Outer_D
        self.ID = ID

        # geometric properties
        self.y_grav_center = 0
        self.z_grav_center = 0
        self.N_vertices = 0
        self.y_vertices = np.array([])
        self.z_vertices = np.array([])

        # Fascicular content
        self.N_ax = 0
        self.fascicles_IDs = []
        self.fascicles = {}

        # Axons objects default parameters
        self.unmyelinated_param = {
            "model": "Rattay_Aberham",
            "dt": 0.001,
            "Nrec": 0,
            "Nsec": 1,
            "Nseg_per_sec": 100,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "v_init": None,
            "T": None,
            "threshold": -40,
        }

        self.myelinated_param = {
            "model": "MRG",
            "dt": 0.001,
            "node_shift": 0,
            "Nseg_per_sec": 3,
            "freq": 100,
            "freq_min": 0,
            "mesh_shape": "plateau_sigmoid",
            "alpha_max": 0.3,
            "d_lambda": 0.1,
            "rec": "nodes",
            "T": None,
            "threshold": -40,
        }

        self.set_axons_parameters(**kwargs)

        # extra-cellular stimulation
        self.extra_stim = None
        self.is_extra_stim = False
        self.added_extra_stims = []
        self.footprints = {}
        self.is_footprinted = False
        self.myelinated_nseg_per_sec = 3
        self.unmyelinated_nseg = 100
        # intra-cellular stimulation
        self.N_intra = 0
        self.intra_stim_position = []
        self.intra_stim_t_start = []
        self.intra_stim_duration = []
        self.intra_stim_amplitude = []
        self.intra_stim_ON = []
        ## recording mechanism
        self.record = False
        self.recorder = None
        self.is_simulated = False
        # Simulation status

    ## save/load methods
    def save(
        self,
        fname="nerve.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        fascicles_context=True,
        save=True,
        _fasc_save=False,
        blacklist=[],
    ):
        """
        Save a fascicle in a json file

        Parameters
        ----------
        fname :             str
            name of the file to save the fascicle
        extracel_context:   bool
            if True, add the extracellular context to the saving
        intracel_context:   bool
            if True, add the intracellular context to the saving
        rec_context:        bool
            if True, add the recording context to the saving
        fascicles_context:  bool
            if True, add the fascicles are fully saved
        blacklist:          list[str]
            key to exclude from saving
        save:               bool
            if false only return the dictionary

        Returns
        -------
        key_dict:           dict
            Pyhton dictionary containing all the fascicle information
        """
        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += ["extra_stim", "footprints", "is_footprinted"]

        if not rec_context:
            bl += ["record", "recorder"]

        if not fascicles_context:
            bl += ["fascicles"]

        to_save = save and MCH.do_master_only_work()
        return super().save(
            fname=fname,
            save=to_save,
            blacklist=bl,
            intracel_context=intracel_context,
            extracel_context=extracel_context,
            rec_context=rec_context,
            _fasc_save=_fasc_save,
        )

    def load(
        self,
        data,
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Load a fascicle configuration from a json file

        Parameters
        ----------
        fname           : str
            path to the json file describing a fascicle
        extracel_context: bool
            if True, load the extracellular context as well
        intracel_context: bool
            if True, load the intracellular context as well
        rec_context: bool
            if True, load the recording context as well
        blacklist       : list[str]
            key to exclude from loading
        """
        if type(data) == str:
            key_dic = json_load(data)
        else:
            key_dic = data

        bl = [i for i in blacklist]
        if not intracel_context:
            bl += [
                "N_intra",
                "intra_stim_position",
                "intra_stim_t_start",
                "intra_stim_duration",
                "intra_stim_amplitude",
                "intra_stim_ON",
            ]

        if not extracel_context:
            bl += [
                "extra_stim",
                "footprints",
                "myelinated_nseg_per_sec",
                "unmyelinated_nseg",
                "is_footprinted",
            ]

        if not rec_context:
            bl += ["record", "recorder"]

        super().load(
            data=key_dic,
            blacklist=bl,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
        )

    ## Nerve property method
    def set_ID(self, ID):
        """
        set the ID of the fascicle

        Parameters
        ----------
        ID : int
            ID number of the fascicle
        """
        self.ID = ID

    def define_length(self, L):
        """
        set the length over the x axis of the fascicle

        Parameters
        ----------
        L   : float
            length of the fascicle in um
        """
        if self.extra_stim is not None or self.N_intra > 0:
            rise_warning(
                "Modifying length of a fascicle with extra or intra cellular context can lead to error"
            )
        self.L = L
        self.set_axons_parameters(unmyelinated_nseg=self.L // 25)
        if self.is_extra_stim:
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )

    ## generate stereotypic Nerve
    def define_circular_contour(self, D, y_c=None, z_c=None, N_vertices=100):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        D           : float
            diameter of the circular fascicle contour, in um
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        N_vertices  : int
            Number of vertice in the compute the contour
        """
        self.type = "Circular"
        self.D = D
        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        self.N_vertices = N_vertices
        theta = np.linspace(-np.pi, np.pi, num=N_vertices)
        self.y_vertices = self.y_grav_center + (D / 2) * np.cos(theta)
        self.z_vertices = self.z_grav_center + (D / 2) * np.sin(theta)
        self.A = np.pi * (D / 2) ** 2
        if self.is_extra_stim:
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )

    def get_circular_contour(self):
        """
        Returns the properties of the fascicle contour considered as a circle (y and z center and diameter)

        Returns
        -------
        D : float
            diameter of the contour, in um. Set to 0 if not applicable
        y : float
            y position of the contour center, in um
        z : float
            z position of the contour center, in um
        """
        y = self.y_grav_center
        z = self.z_grav_center
        if self.y_vertices == np.array([]) and self.D is None:
            D = 0
        elif self.D is not None:
            D = self.D
        else:
            y = (np.amax(self.y_vertices) - np.amin(self.y_vertices)) / 2
            z = (np.amax(self.z_vertices) - np.amin(self.z_vertices)) / 2
            D = np.abs(np.amax(self.y_vertices) - np.amin(self.y_vertices))
        return D, y, z

    def fit_circular_contour(self, y_c=None, z_c=None, Delta=20, N_vertices=100):
        """
        Define a circular countour to the fascicle

        Parameters
        ----------
        y_c         : float
            y coordinate of the circular contour center, in um
        z_c         : float
            z coordinate of the circular contour center, in um
        Delta       : float
            distance between farest axon and contour, in um
        N_vertices  : int
            Number of vertice in the compute the contour
        """
        N_fasc = len(self.fascicles)
        D = 2 * Delta

        if y_c is not None:
            self.y_grav_center = y_c
        if z_c is not None:
            self.z_grav_center = z_c
        if N_fasc == 0:
            pass_info("No fascicles to fit fascicul diameter set to " + str(D) + "um")
        else:
            for fasc in self.fascicles.values():
                dist_max = (
                    fasc.D / 2
                    + (
                        (self.y_grav_center - fasc.y_grav_center) ** 2
                        + (self.z_grav_center - fasc.z_grav_center) ** 2
                    )
                    ** 0.5
                )
                D = max(D, 2 * (dist_max + Delta))
        self.define_circular_contour(D, y_c=None, z_c=None, N_vertices=N_vertices)

    def define_ellipsoid_contour(self, a, b, y_c=0, z_c=0, rotate=0):
        """
        Define ellipsoidal contour
        """
        pass

    ## generate Fascicle from histology
    def import_contour(self, smthing_else):
        """
        Define contour from a file
        """
        pass

    ## Fascicles handeling methods
    def get_fascicles(self, ID_only=False):
        if ID_only:
            return self.fascicles_IDs
        else:
            return self.fascicles

    def add_fascicle(self, fascicle, ID=None, y=None, z=None, **kwargs):
        """
        Add a fascicle to the list of fascicles

        Parameters
        ----------
        fascicle : object
            fascicle to add to the nerve struture
        """
        fasc = load_any(fascicle, **kwargs)
        if not is_fascicle(fasc):
            rise_warning(
                "Only fascile (nrv object, dict or file) can be added with add fascicle method: nothing added"
            )
        else:
            if ID is not None:
                fasc.set_ID(ID)
            if y is None:
                y = fasc.y_grav_center
            if z is None:
                z = fasc.z_grav_center
            fasc.translate_fascicle(y - fasc.y_grav_center, z - fasc.z_grav_center)
            if self.__check_fascicle_overlap(fasc):
                rise_warning(
                    "fascicles overlap:  fasicle " + str(fasc.ID) + " cannot be added"
                )
            else:
                if fasc.ID in self.fascicles_IDs:
                    pass_info(
                        "Fascicle "
                        + str(fasc.ID)
                        + " already in the nerve: will be replace"
                    )
                    self.fascicles[fasc.ID]
                else:
                    self.fascicles_IDs += [fasc.ID]
                    self.fascicles[fasc.ID] = fasc

                self.__merge_fascicular_context(self.fascicles[fasc.ID])
                self.fascicles[fasc.ID].define_length(self.L)

    def __check_fascicle_overlap(self, fasc):
        for fasc_i in self.fascicles.values():
            dist_yz = (fasc.y_grav_center - fasc_i.y_grav_center) ** 2 + (
                fasc.z_grav_center - fasc_i.z_grav_center
            ) ** 2
            len_min_yz = ((fasc.D + fasc_i.D) / 2) ** 2
            if dist_yz < len_min_yz:
                return True
        return False

    def __merge_fascicular_context(self, fasc: fascicle):
        if self.is_extra_stim:
            self.extra_stim.reshape_fascicle(
                fasc.D, fasc.y_grav_center, fasc.z_grav_center, fasc.ID
            )
        if fasc.extra_stim is not None:
            self.attach_extracellular_stimulation(fasc.extra_stim)

        if fasc.recorder is not None:
            self.attach_extracellular_recorder(fasc.recorder)

        fasc.clear_context(intracel_context=False)

    ## Axons handeling method
    def set_axons_parameters(
        self, unmyelinated_only=False, myelinated_only=False, **kwargs
    ):
        """
        set parameters of axons in the fascicle

        Parameters
        ----------
        unmyelinated_only:      bool
            if true modification only done for unmyelinated axons parameters, by default False
        myelinated_only:        bool
            if true modification only done for myelinated axons parameters, by default False
        kwargs:
            parameters to modify (see myelinated and unmyelinated)
        """
        ## Standard keys
        for key in kwargs:
            if "model" in key:
                if not myelinated_only and kwargs[key] in unmyelinated_models:
                    self.unmyelinated_param["model"] = kwargs[key]
                elif not unmyelinated_only and kwargs[key] in myelinated_models:
                    self.myelinated_param["model"] = kwargs[key]
                else:
                    rise_warning(kwargs[key], " is not an implemented axon model")
            else:
                if not myelinated_only and key in self.unmyelinated_param:
                    self.unmyelinated_param["model"] = kwargs[key]
                if not unmyelinated_only and key in self.myelinated_param:
                    self.myelinated_param["model"] = kwargs[key]

        ## Specific keys
        if "unmyelinated_nseg" in kwargs:
            self.unmyelinated_param["Nseg_per_sec"] = kwargs["unmyelinated_nseg"]
        if "myelinated_nseg_per_sec" in kwargs:
            self.myelinated_param["Nseg_per_sec"] = kwargs["myelinated_nseg_per_sec"]
        if "Adelta_limit" in kwargs:
            self.Adelta_limit = kwargs["self.Adelta_limit"]

        for fasc in self.fascicles.values():
            fasc.set_axons_parameters(unmyelinated_only=True, **self.unmyelinated_param)
            fasc.set_axons_parameters(myelinated_only=True, **self.myelinated_param)

    def get_axons_parameters(self, unmyelinated_only=False, myelinated_only=False):
        """
        get parameters of axons in the nerve

        Parameters
        ----------
        unmyelinated_only:      bool
            modification will only
        myelinated_only:        bool
            modification will only
        """
        if unmyelinated_only:
            return self.unmyelinated_param
        if myelinated_only:
            return self.myelinated_param
        return self.unmyelinated_param, self.myelinated_param

    ## Representation methods
    def plot(
        self, fig, axes, contour_color="k", myel_color="r", unmyel_color="b", num=False
    ):
        """
        plot the nerve in the Y-Z plane (transverse section)

        Parameters
        ----------
        fig     : matplotlib.figure
            figure to display the fascicle
        axes    : matplotlib.axes
            axes of the figure to display the fascicle
        contour_color   : str
            matplotlib color string applied to the contour. Black by default
        myel_color      : str
            matplotlib color string applied to the myelinated axons. Red by default
        unmyel_color    : str
            matplotlib color string applied to the myelinated axons. Blue by default
        num             : bool
            if True, the index of each axon is displayed on top of the circle
        """
        if MCH.do_master_only_work():
            ## plot contour
            axes.plot(
                self.y_vertices, self.z_vertices, linewidth=4, color=contour_color
            )
            for i in self.fascicles:
                fasc = self.fascicles[i]
                fasc.plot(
                    fig=fig,
                    axes=axes,
                    contour_color=contour_color,
                    myel_color=myel_color,
                    unmyel_color=unmyel_color,
                    num=num,
                )

    ## Context handeling methods
    def clear_context(
        self, extracel_context=True, intracel_context=True, rec_context=True
    ):
        """
        Clear all stimulation and recording mecanism
        """
        if extracel_context:
            self.L = None
            # extra-cellular stimulation
            self.extra_stim = None
            self.added_extra_stims = None
            self.footprints = {}
            self.is_footprinted = False
        if intracel_context:
            # intra-cellular stimulation
            self.N_intra = 0
            self.intra_stim_position = []
            self.intra_stim_t_start = []
            self.intra_stim_duration = []
            self.intra_stim_amplitude = []
            self.intra_stim_ON = []
        if rec_context:
            ## recording mechanism
            self.record = False
            self.recorder = None
        self.is_simulated = False

    # Intra cellular

    def insert_I_Clamp(self, position, t_start, duration, amplitude, ax_list=None):
        """
        Insert a IC clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the fascicle. Note that all thin myelinated and myelinated
            will be stimulated in the nearest node of Ranvier around the clamp specified position
        t_start     : float
            starting time, in ms
        duration    : float
            duration of the pulse, in ms
        amplitude   : float
            amplitude of the pulse (nA)
        ax_list     : list, array, np.array
            list of axons to insert the clamp on, if None, all axons are stimulated
        """
        for fasc in self.fascicles.values():
            fasc.insert_I_Clamp(
                position=position,
                t_start=t_start,
                duration=duration,
                amplitude=amplitude,
                ax_list=ax_list,
            )
        self.N_intra += 1

    # Extracellular

    def attach_extracellular_stimulation(self, stimulation: FEM_stimulation):
        """
        attach a extracellular context of simulation for an axon.

        Parameters
        ----------
            stimulation  : stimulation object, see Extracellular.stimulation help for more details
        """
        if not self.is_extra_stim:
            self.extra_stim = FEM_stimulation(
                endo_mat=stimulation.endoneurium,
                peri_mat=stimulation.perineurium,
                epi_mat=stimulation.epineurium,
                ext_mat=stimulation.external_material,
            )

            self.extra_stim.reshape_outerBox(self.Outer_D)
            self.extra_stim.reshape_nerve(
                Nerve_D=self.D,
                Length=self.L,
                y_c=self.y_grav_center,
                z_c=self.z_grav_center,
            )
            self.extra_stim.remove_fascicles()
            for fasc in self.fascicles.values():
                self.extra_stim.reshape_fascicle(
                    Fascicle_D=fasc.D,
                    y_c=fasc.y_grav_center,
                    z_c=fasc.z_grav_center,
                    ID=fasc.ID,
                )
            self.is_extra_stim = True

        for i, elec in enumerate(stimulation.electrodes):
            self.extra_stim.add_electrode(elec, stimulation.stimuli[i])

    def change_stimulus_from_elecrode(self, ID_elec, stimulus):
        """
        Change the stimulus of the ID_elec electrods

        Parameters:
        ----------
            ID_elec  : int
                ID of the electrode which should be changed
            stimulus  : stimulus
                New stimulus to set
        """
        if self.extra_stim is not None:
            self.extra_stim.change_stimulus_from_elecrode(ID_elec, stimulus)
        else:
            rise_warning("Cannot be changed: no extrastim in the nerve")

    # RECORDING MECHANIMS
    def attach_extracellular_recorder(self, rec: recorder):
        """
        attach an extracellular recorder to the axon

        Parameters
        ----------
        rec     : recorder object
            see Recording.recorder help for more details
        """
        if is_recorder(rec):
            if self.recorder is None:
                self.recorder = rec
            else:
                for rec_p in rec.recording_points:
                    self.recorder.add_recording_point(rec_p)
            self.record = True

    def shut_recorder_down(self):
        """
        Shuts down the recorder locally
        """
        self.record = False
        for fasc in self.fascicles.values():
            fasc.shut_recorder_down()

    def __set_fascicles_context(self):
        """ """
        for fasc in self.fascicles.values():
            if self.extra_stim is not None:
                fasc.attach_extracellular_stimulation(self.extra_stim)
            if self.recorder is not None:
                fasc.attach_extracellular_recorder(self.recorder)

    def __set_fascicles_simulation_parameters(self):
        for fasc in self.fascicles.values():
            fasc.t_sim=self.t_sim
            fasc.record_V_mem=self.record_V_mem
            fasc.record_I_mem=self.record_I_mem,
            fasc.record_I_ions=self.record_I_ions
            fasc.record_g_mem=self.record_g_mem
            fasc.record_g_ions=self.record_g_ions
            fasc.record_particles=self.record_particles
            fasc.postproc_script=self.postproc_script


    def compute_electrodes_footprints(self, **kwargs):
        """
        compute electrodes footprints
        """
        if not self.is_footprinted:
            self.__set_fascicles_context()
            self.set_axons_parameters(**kwargs)

            if self.is_extra_stim:
                mp_computation = False
                if is_FEM_extra_stim(self.extra_stim):
                    mp_computation = (
                        self.extra_stim.fenics and self.extra_stim.model.is_multi_proc
                    )
                    if MCH.do_master_only_work() or mp_computation:
                        self.extra_stim.run_model()
                for fasc in self.fascicles.values():
                    fasc.compute_electrodes_footprints()
            self.is_footprinted = True

    ## SIMULATION
    def simulate(
        self,
        **kwargs,
    ):
        """
        Simulate the nerve with the proposed extracellular context. Top level method for large
        scale neural simulation.

        Parameters
        ----------

        Warning
        -------
        calling this method can result in long processing time, even with large computational ressources.
        Keep aware of what is really implemented, ensure configuration and simulation protocol is correctly designed.
        """
        pass_info("Starting nerve simulation")
        nerve_sim = super().simulate(**kwargs)
        self.__set_fascicles_simulation_parameters()
        if not MCH.do_master_only_work():
            nerve_sim = {}
        folder_name = self.save_path + "Nerve_" + str(self.ID) + "/"
        if MCH.do_master_only_work():
            create_folder(folder_name)
            config_filename = folder_name + "/00_Nerve_config.json"
            self.save(config_filename, fascicles_context=False)
        else:
            pass
        # run FEM model
        pass_info("...computing electrodes footprint")
        self.compute_electrodes_footprints(**kwargs)
        synchronize_processes()
        # Simulate all fascicles
        fasc_kwargs = kwargs
        fasc_kwargs["save_path"] = folder_name
        for fasc in self.fascicles.values():
            pass_info("...simulating fascicle " + str(fasc.ID))
            nerve_sim["fascicle"+str(fasc.ID)] = fasc.simulate(
                in_nerve=True,
                **fasc_kwargs,
            )
        self.is_simulated = True
        return nerve_sim



from ....backend.log_interface import rise_warning
from ....nmod.fascicles import *
from ....utils.units import *
from ...electrodes import *
from .NerveMshCreator import *


def mesh_from_electrode(
    elec,
    mesh=None,
    Length=10000,
    Outer_D=5,
    Nerve_D=4000,
    y_c=0,
    z_c=0,
    res="default",
):
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    if not is_FEM_electrode(elec):
        rise_warning("Only FEM electrode can be added to a mesh")
    else:
        if mesh is None:
            mesh = NerveMshCreator(
                Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c
            )
        elec.parameter_model(mesh, res=res)
    return mesh


def mesh_from_fascicle(
    fascicle,
    mesh=None,
    Length=10000,
    Outer_D=5,
    Nerve_D=4000,
    y_c=0,
    z_c=0,
    res_nerve="default",
    res_fasc="default",
    res_ax="default",
    res_elec="default",
):
    """
    returns the corresponding mesh from a nrv.facsicle
    """
    if mesh is None:
        mesh = NerveMshCreator(
            Length=Length, Outer_D=Outer_D, Nerve_D=Nerve_D, y_c=y_c, z_c=z_c
        )
        if res_nerve != "default":
            mesh.reshape_nerve(res=res_nerve)

    mesh.reshape_fascicle(
        D=fascicle.D,
        y_c=fascicle.y_grav_center,
        z_c=fascicle.z_grav_center,
        res=res_fasc,
    )

    for i_ax in range(fascicle.N):
        ax_d = round(fascicle.axons_diameter[i_ax], 3)
        ax_y_c = round(fascicle.axons_y[i_ax], 3)
        ax_z_c = round(fascicle.axons_z[i_ax], 3)
        if isinstance(res_ax, str) and res_ax != "default":
            res_ax = eval(str(fascicle.axons_diameter[i_ax]) + res_ax)
        mesh.reshape_axon(D=ax_d, y_c=ax_y_c, z_c=ax_z_c, res=res_ax)

    if fascicle.extra_stim is not None:
        N_contact = 0
        if np.iterable(res_elec) and not isinstance(res_elec, str):
            if len(res_elec) == len(fascicle.extra_stim.electrodes):
                res = res_elec
            else:
                rise_warning(
                    "length of res_elec and fascicle.extra_stim.electrodes does not match",
                    "only first value kept",
                )
                res = [res_elec[0] for k in range(len(fascicle.extra_stim.electrodes))]
        else:
            res = [res_elec for k in range(len(fascicle.extra_stim.electrodes))]
        for i_elec, elec in enumerate(fascicle.extra_stim.electrodes):
            if (not elec.is_multipolar) or N_contact <= 0:
                mesh = mesh_from_electrode(elec, mesh=mesh, res=res[i_elec])
                if elec.is_multipolar:
                    N_contact = elec.N_contact
            else:
                N_contact -= 1
    return mesh


def mesh_from_nerve(nerve, mesh=None):
    rise_warning("not implemented yet")



"""
log_interface handeling
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import logging
import os
import sys

from icecream import ic

from .MCore import MCH
from .parameters import parameters

from pyswarms.utils import Reporter




class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


dir_path = os.environ["NRVPATH"] + "/_misc"

rep_nrv = Reporter(log_path=dir_path + "/log/NRV.log")
rep_nrv._default_config['handlers']['default']['level']= 'CRITICAL'
rep_nrv._default_config['formatters']['standard']['format']= "%(asctime)s - %(message)s"
rep_nrv._load_defaults()


"""
!! ISSUE:
Compatibility between logging and pyswarms when pyswarms is imported (not used necessarily)
all messages log are also printed

!! TEMPORARY SOLUTION
Using a pyswarms reporter to log messages 
see old code commented 

logger = logging.getLogger()

# logging config
logging.basicConfig(
    filename=dir_path + "/log/NRV.log",
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
)"""


def rise_error(*args, out=1, **kwargs):
    """
    Rises and error to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as error message
    out     : int
        programm exit value, must be strictly positive. By default equal to 1
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]

    message = ""
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            rep_nrv.log(message, lvl=logging.ERROR)
            #logger.error(message, lvl=logging.ERROR)
        if verbose and parameters.VERBOSITY_LEVEL >= 1:
            print(bcolors.FAIL + "NRV ERROR: " + message + bcolors.ENDC)
    else:
        err = (
            "NRV ERROR: "
            + message
            + "\n encountered in process "
            + str(MCH.rank)
            + " out of "
            + str(MCH.size)
        )
        if parameters.LOG_Status:
            rep_nrv.log(err, lvl=logging.ERROR)
            #logger.error(err, lvl=logging.ERROR)
        if parameters.VERBOSITY_LEVEL >= 1:
            print(err)
            sys.stdout.flush()
    if out == 0:
        out = 1
    sys.exit(out)


def rise_warning(*args, abort=False, **kwargs):
    """
    Rises a warning to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing. This function can exit the programm.

    Parameters
    ----------
    *args   :
        anything to pass as warning message
    abort   : boolean
        if true, the programm exits with the value O (no error) and further computation are avoided,
        by default set to False.
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            rep_nrv.log("NRV DEBUG: " + message,lvl=logging.DEBUG)
            #logger.warning("NRV WARNING: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 2:
            print(bcolors.WARNING + "NRV WARNING: " + message + bcolors.ENDC)
    else:
        war = (
            "NRV WARNING: "
            + message
            + "\n encountered in process "
            + str(MCH.rank)
            + " out of "
            + str(MCH.size)
        )
        if parameters.LOG_Status:
            rep_nrv.log(war,lvl=logging.DEBUG)
            #logger.warning(war)
        if MCH.do_master_only_work() and parameters.VERBOSITY_LEVEL >= 2:
            print(bcolors.WARNING + war + bcolors.ENDC)
            sys.stdout.flush()
        elif parameters.VERBOSITY_LEVEL >= 4:
            print(bcolors.WARNING + war + bcolors.ENDC)
            sys.stdout.flush()
    if abort:
        sys.exit(0)


def pass_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status:
            rep_nrv.log("NRV DEBUG: " + message,lvl=logging.INFO)
            #logger.info("NRV DEBUG: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 3:
            print("NRV INFO: " + message)
    else:
        inf = (
            "NRV INFO: "
            + message
            + "\n from process "
            + str(MCH.rank)
            + " out of "
            + str(MCH.size)
        )
        if parameters.LOG_Status:
            rep_nrv.log(inf,lvl=logging.INFO)
            #logger.info(inf)
        if parameters.VERBOSITY_LEVEL >= 4:
            print(inf)
            sys.stdout.flush()


def pass_debug_info(*args, **kwargs):
    """
    Pass an info to the log and to the prompt for the master process (process ID 0) in case of
    parallel computing.

    Parameters
    ----------
    *args   :
        anything to pass as info
    """
    verbose = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    message = ""
    for arg in args:
        message += str(arg)
    if MCH.is_alone():
        if parameters.LOG_Status and parameters.VERBOSITY_LEVEL >= 4:
            rep_nrv.log("NRV DEBUG: " + message,lvl=logging.DEBUG)
            #logger.info("NRV DEBUG: " + message)
        if verbose and parameters.VERBOSITY_LEVEL >= 4:
            print("NRV DEBUG: " + message)
    else:
        inf = (
            "NRV DEBUG: "
            + message
            + "\n from process "
            + str(MCH.rank)
            + " out of "
            + str(MCH.size)
        )
        if parameters.LOG_Status:
            rep_nrv.log(inf,lvl=logging.DEBUG)
            #logger.info(inf)
        if parameters.VERBOSITY_LEVEL >= 4:
            print(inf)
            sys.stdout.flush()


def progression_popup(current, max_iter, begin_message="", end_message="", endl=""):
    """
    Displays the progression on prompt, for single process only, nothing saved to the log

    Parameters
    ----------
    current         : int
        current iteration
    max_iter        : int
        maximum iteration number
    begin_message   : str
        additional message placed at the start of the progression line
    end_message     : str
        additional message placed after iteration and max at the end of the progression line
    endl            : str
        line termination, by default empty string
    """
    if MCH.is_alone():
        print(
            begin_message + f"{current + 1}" + "/" + str(max_iter) + end_message,
            end=endl,
        )
        if current == max_iter - 1:
            print("\n")


def prompt_debug(*args):
    """
    outputs for debug, simple redirection to icecream without the installation (see icecream help).
    Could be interesting to redirect also to the log...

    Parameters
    ----------
    *args   :
        anything to pass to icecrean
    """
    ic(*args)



"""
NRV-fenics_materials
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
import os

import numpy as np
from dolfinx.fem import Constant, Function, FunctionSpace
from petsc4py.PETSc import ScalarType
from ufl import as_tensor

from ....backend.file_handler import json_dump, rmv_ext
from ....backend.log_interface import rise_warning
from ....utils.nrv_function import nrv_interp
from ...materials import is_mat, load_material, material

# enable faulthandler to ease "segmentation faults" debug
faulthandler.enable()

# get the built-in material librairy
dir_path = os.environ["NRVPATH"] + "/_misc"

###############
## Functions ##
###############


def is_fen_mat(mat):
    """
    check if an object is a fenics_material, return True if yes, else False

    Parameters
    ----------
    mat : object
        object to test

    Returns
    -------
    bool
        True it the type is a material object
    """
    return isinstance(mat, fenics_material)


def mat_from_interp(
    X, Y, kind="linear", dx=0.01, interpolator=None, dxdy=None, scale=None, columns=0
):
    """
    Return a fenics material with a conductivity space function define as the nrv_interp of X and Y
    (see nrv_function for more details)

    Parameters
    ----------
    f_material  : str
        either material name if the material is in the NRV2 Librairy or path to the corresponding
        .mat material file
    kwargs      : dict
        interpolation k arguments (cf ....utils.nrv_function.nrv_interp)

    Returns
    -------
    mat_obj     : fenics_material
        Fenics material with conductivity function defined from the interpolation
    """
    sigma_func = nrv_interp(
        X,
        Y,
        kind=kind,
        dx=dx,
        interpolator=interpolator,
        dxdy=dxdy,
        scale=scale,
        columns=columns,
    )

    # creat material instance
    mat_obj = fenics_material()
    mat_obj.set_conductivity_function(sigma_func)
    return mat_obj


def mat_from_csv(f_material, **kwargs):
    """
    Return a fenics material with a conductivity space function define as the nrv_interp from the
    value of a .csv file

    Parameters
    ----------
    f_material  : str
        name of the .csv file

    Returns
    -------
    mat_obj     : fenics_material
        Fenics material with conductivity function defined from the interpolation from the
        value of f_material file
    """
    fname = rmv_ext(f_material) + ".csv"

    data = np.loadtxt(fname, delimiter=",")
    X = data[0]
    Y = data[1]

    return mat_from_interp(X, Y, **kwargs)


def load_fenics_material(X, **kwargs):
    """
    Return fenics material from an object X

    Parameters
    ----------
    X   : objects
        if material or fenics material returning the corresponding fenics material
        if float returns an isotropic fenics material with X as conductivity value
        if str finishing with .csv use mat_from_csv with kwargs
        if
        else return None
    """
    mat = None
    if is_fen_mat(X):
        mat = X
    elif is_mat(X):
        mat = fenics_material(X)
    elif isinstance(X, str):
        if ".csv" in X:
            mat = mat_from_csv(X, **kwargs)
        else:
            mat = fenics_material(load_material(X))
    elif isinstance(X, (float, int)):
        mat = fenics_material()
        mat.set_isotropic_conductivity(X)
    elif np.iterable(X):
        if len(X) == 2:
            mat = mat_from_interp(X[0], X[1], **kwargs)
        elif len(X) == 3:
            mat = fenics_material()
            mat.set_anisotropic_conductivity(X[0], X[1], X[2])
    return mat


####################
## material class ##
####################
class fenics_material(material):
    """
    a class for material material more suited for the FEM solving with fenics
    Inherit from material class. see material for further detail
    """

    def __init__(self, mat=None):
        """
        initialisation of the fenics_material
        mat     :material
            generate the fenics material from mat attribute
        """
        super().__init__()
        self.is_func = False
        self.sigma_func = None
        self.sigma_fen = None
        self.elem = ("Discontinuous Lagrange", 1)
        self.UN = 1

        if is_mat(mat):
            self.name = mat.name
            self.source = mat.source
            self.isotrop_cond = mat.isotrop_cond
            self.sigma = mat.sigma
            self.sigma_xx = mat.sigma_xx
            self.sigma_yy = mat.sigma_yy
            self.sigma_zz = mat.sigma_zz

    ## Save and Load mehtods
    def save(self, save=False, fname="Fenics_model.json", blacklist=[], **kwargs):
        """
        Return material as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "material.json"

        Returns
        -------
        mat_dic : dict
            dictionary containing all information
        """
        bl = [i for i in blacklist]
        bl += ["sigma_fen"]
        return super().save(save=save, fname=fname, blacklist=bl, **kwargs)

    def save_fenics_material(self, save=False, fname="fenics_material.json"):
        rise_warning("save_fenics_material is a deprecated method use save")
        self.save(save=save, fname=fname)

    def load_fenics_material(self, data="fenics_material.json"):
        rise_warning("load_fenics_material is a deprecated method use load")
        self.load(data=data)

    def load_from_mat(self, mat):
        if is_mat(mat):
            self.name = mat.name
            self.source = mat.source
            self.isotrop_cond = mat.isotrop_cond
            self.sigma = mat.sigma
            self.sigma_xx = mat.sigma_xx
            self.sigma_yy = mat.sigma_yy
            self.sigma_zz = mat.sigma_zz
        if is_fen_mat(mat):
            self.is_func = mat.is_func
            self.sigma_func = mat.sigma_func

        else:
            rise_warning("Not an material nothing is load")
        self.update_fenics_sigma()

    def set_conductivity_function(self, sigma_func):
        """
        set the conductivity space function for an anisotropic material

        Parameters
        ----------
        sigma_xx    : func(X : array([3, N]))-> array([N])
            conductivity function in 3D space

        """
        self.isotrop_cond = False
        self.is_func = True
        self.sigma_func = sigma_func

    def is_function_defined(self):
        """
        check that the material conductivity is define as a function

        Returns
        -------
        bool    :
            True if the per
        """
        return self.is_func

    def get_fenics_sigma(
        self, domain, elem=("Discontinuous Lagrange", 1), UN=1, id=None
    ):
        """
        Returns fenicsx compatible sigma
        """
        self.update_fenics_sigma(domain, elem=elem, UN=UN, id=id)
        return self.sigma_fen

    def update_fenics_sigma(self, domain=None, elem=None, UN=None, id=None):
        if id is None:
            name = ""
        else:
            name = "_" + str(id)
        if elem is not None:
            self.elem = elem
        if UN is not None:
            self.UN = UN
        if self.sigma_fen is None:
            if self.is_isotropic():
                self.sigma_fen = Constant(domain, ScalarType(self.sigma * self.UN))
            elif not self.is_func:
                self.sigma_fen = as_tensor(
                    [
                        [self.sigma_xx * self.UN, 0, 0],
                        [0, self.sigma_yy * self.UN, 0],
                        [0, 0, self.sigma_zz * self.UN],
                    ]
                )
            else:
                unit = Constant(domain, ScalarType(self.UN))
                Q = FunctionSpace(domain, self.elem)
                self.sigma_fen = Function(Q)
                self.sigma_fen.interpolate(self.sigma_func)
                self.sigma_fen.name = "f" + name
                self.sigma_fen = self.sigma_fen * unit
        else:
            if self.is_isotropic():
                self.sigma_fen.value = self.sigma * self.UN

            elif not self.is_func:
                self.sigma_fen = as_tensor(
                    [
                        [self.sigma_xx * self.UN, 0, 0],
                        [0, self.sigma_yy * self.UN, 0],
                        [0, 0, self.sigma_zz * self.UN],
                    ]
                )
            else:
                self.sigma_fen.ufl_operands[0].interpolate(self.sigma_func)
                self.sigma_fen.ufl_operands[1].value = self.UN
